# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2015,2017
"""
SPL Python primitive operators.

********
Overview
********

SPL primitive operators that call a Python function or
class methods are created by decorators provided by this module.

The name of the function or callable class becomes the name of the
operator.

A decorated function is a stateless operator while a decorated class
is an optionally stateful operator.

These are the supported decorators that create an SPL operator:

    * :py:class:`@spl.source <source>` - Creates a source operator that produces tuples.
    * :py:class:`@spl.filter <filter>` - Creates a operator that filters tuples.
    * :py:class:`@spl.map <map>` - Creates a operator that maps input tuples to output tuples.
    * :py:class:`@spl.for_each <for_each>` - Creates a operator that terminates a stream processing each tuple.
    * :py:class:`@spl.primitive_operator <primitive_operator>` - Creates an SPL primitive operator that has an arbitrary number of input and output ports.

Decorated functions and classes must be located in the directory
``opt/python/streams`` in the SPL toolkit. Each module in that directory
will be inspected for operators during extraction. Each module defines
the SPL namespace for its operators by the function ``spl_namespace``,
for example::

    from streamsx.spl import spl

    def spl_namespace():
        return 'com.example.ops'

    @spl.map()
    def Pass(*tuple_):
        return tuple_

creates a pass-through operator ``com.example.ops::Pass``.

SPL primitive operators are created by executing the extraction script :ref:`spl-py-extract` against the SPL toolkit. Once created the operators become part
of the toolkit and may be used like any other SPL operator.

*******************************
Python classes as SPL operators
*******************************

Overview
========

Decorating a Python class creates a stateful SPL operator
where the instance fields of the class are the operator's state. An instance
of the class is created when the SPL operator invocation is initialized
at SPL runtime. The instance of the Python class is private to the SPL
operator and is maintained for the lifetime of the operator.

If the class has instance fields then they are the state of the
operator and are private to each invocation of the operator.

If the `__init__` method has parameters beyond the first
`self` parameter then they are mapped to operator parameters.
Any parameter that has a default value becomes an optional parameter
to the SPL operator. Parameters of the form `\*args` and `\*\*kwargs`
are not supported.

.. warning::
    Parameter names must be valid SPL identifers,
    SPL identifiers start with an ASCII letter or underscore,
    followed by ASCII letters, digits, or underscores.
    The name also must not be an SPL keyword.

    Parameter names ``suppress`` and ``include`` are reserved.

The value of the operator parameters at SPL operator invocation are passed
to the `__init__` method. This is equivalent to creating an instance
of the class passing the operator parameters into the constructor.

For example, with this decorated class producing an SPL source
operator::

    @spl.source()
    class Range(object):
      def __init__(self, stop, start=0):
        self.start = start
        self.stop = stop

      def __iter__(self):
          return zip(range(self.start, self.stop))

The SPL operator `Range` has two parameters, `stop` is mandatory and `start` is optional, defaulting to zero. Thus the SPL operator may be invoked as::

    // Produces the sequence of values from 0 to 99
    //
    // Creates an instance of the Python class
    // Range using Range(100)
    //
    stream<int32 seq> R = Range() {
      param
        stop: 100;
    }

or both operator parameters can be set::

    // Produces the sequence of values from 50 to 74
    //
    // Creates an instance of the Python class
    // Range using Range(75, 50)
    //
    stream<int32 seq> R = Range() {
      param
        start: 50;
        stop: 75;
    }

Operator state
==============

Use of a class allows the operator to be stateful by maintaining state in instance
attributes across invocations (tuple processing).

When the operator is in a consistent region or checkpointing then it is serialized using `dill`. The default serialization may be modified by using the standard Python pickle mechanism of ``__getstate__`` and ``__setstate__``. This is required if the state includes objects that cannot be serialized, for example file descriptors. For details see See https://docs.python.org/3.5/library/pickle.html#handling-stateful-objects .

If the class has ``__enter__`` and ``__exit__`` context manager methods then ``__enter__`` is called after the instance has been deserialized by `dill`. Thus ``__enter__`` is used to recreate runtime objects that cannot be serialized such as open files or sockets.

Operator initialization & shutdown
==================================

Execution of an instance for an operator effectively run in a context manager so that an instance's ``__enter__``
method is called when the processing element containing the operator is initialized
and its ``__exit__`` method called when the processing element is stopped. To take advantage of this
the class must define both ``__enter__`` and ``__exit__`` methods.

.. note::
    Initialization such as opening files should be in ``__enter__``
    in order to support stateful operator restart & checkpointing.

Example of using ``__enter__`` and ``__exit__`` to open and close a file::

    import streamsx.ec as ec

    @spl.map()
    class Sentiment(object):
        def __init__(self, name):
            self.name = name
            self.file = None

        def __enter__(self):
            self.file = open(self.name, 'r')

        def __exit__(self, exc_type, exc_value, traceback):
            if self.file is not None:
                self.file.close()

        def __call__(self):
            pass

When an instance defines a valid ``__exit__`` method then it will be called with an exception when:

 * the instance raises an exception during processing of a tuple
 * a data conversion exception is raised converting a Python value to an SPL tuple or attribute

If ``__exit__`` returns a true value then the exception is suppressed and processing continues, otherwise the enclosing processing element will be terminated.

Application log and trace
=========================

IBM Streams provides application trace and log services which are
accesible through standard Python loggers from the `logging` module.

See :ref:`streams_app_log_trc`.

*********************************
Python functions as SPL operators
*********************************

Decorating a Python function creates a stateless SPL operator.
In SPL terms this is similar to an SPL Custom operator, where
the code in the Python function is the custom code. For
operators with input ports the function is called for each
input tuple, passing a Python representation of the SPL input tuple.
For an SPL source operator the function is called to obtain an iterable
whose contents will be submitted to the output stream as SPL tuples.

Operator parameters are not supported.

An example SPL sink operator that prints each input SPL tuple after
its conversion to a Python tuple::

    @spl.for_each()
    def PrintTuple(*tuple_):
        "Print each tuple to standard out."
         print(tuple_, flush=True)

.. _spl-tuple-to-python:

*******************************
Processing SPL tuples in Python
*******************************

SPL tuples are converted to Python objects and passed to a decorated callable.

Overview
========

For each SPL tuple arriving at an input port a Python function is called with
the SPL tuple converted to Python values suitable for the function call.
How the tuple is passed is defined by the tuple passing style.

Tuple Passing Styles
====================

An input tuple can be passed to Python function using a number of different styles:
 * *dictionary*
 * *tuple*
 * *attributes by name* **not yet implemented**
 * *attributes by position*

Dictionary
----------

Passing the SPL tuple as a Python dictionary is flexible
and makes the operator independent of any schema.
A disadvantage is the reduction in code readability
for Python function by not having formal parameters,
though getters such as ``tuple['id']`` mitigate that to some extent.
If the function is general purpose and can derive meaning
from the keys that are the attribute names then ``**kwargs`` can be useful.

When the only function parameter is ``**kwargs``
(e.g. ``def myfunc(**tuple_):``) then the passing style is *dictionary*.

All of the attributes are passed in the dictionary
using the SPL schema attribute name as the key.

Tuple
-----

Passing the SPL tuple as a Python tuple is flexible
and makes the operator independent of any schema
but is brittle to changes in the SPL schema.
Another disadvantage is the reduction in code readability
for Python function by not having formal parameters.
However if the function is general purpose and independent
of the tuple contents ``*args`` can be useful.

When the only function parameter is ``*args``
(e.g. ``def myfunc(*tuple_):``) then the passing style is *tuple*.

All of the attributes are passed as a Python tuple
with the order of values matching the order of the SPL schema.

Attributes by name
------------------
(**not yet implemented**)

Passing attributes by name can be robust against changes
in the SPL scheme, e.g. additional attributes being added in
the middle of the schema, but does require that the SPL schema
has matching attribute names.

When *attributes by name* is used then SPL tuple attributes
are passed to the function by name for formal parameters.
Order of the attributes and parameters need not match.
This is supported for function parameters of
kind ``POSITIONAL_OR_KEYWORD`` and ``KEYWORD_ONLY``.

If the function signature also contains a parameter of the form
``**kwargs`` (``VAR_KEYWORD``) then any attributes not bound to
formal parameters are passed in its dictionary using the
SPL schema attribute name as the key.

If the function signature also contains an arbitrary argument
list ``*args`` then any attributes not bound to formal parameters
or to ``**kwargs`` are passed in order of the SPL schema.

If there are only formal parameters any non-bound attributes
are not passed into the function.

Attributes by position
----------------------

Passing attributes by position allows the SPL operator to
be independent of the SPL schema but is brittle to
changes in the SPL schema. For example a function expecting
an identifier and a sensor reading as the first two attributes
would break if an attribute representing region was added as
the first SPL attribute.

When *attributes by position* is used then SPL tuple attributes are
passed to the function by position for formal parameters.
The first SPL attribute in the tuple is passed as the first parameter.
This is supported for function parameters of kind `POSITIONAL_OR_KEYWORD`.

If the function signature also contains an arbitrary argument
list `\*args` (`VAR_POSITIONAL`) then any attributes not bound
to formal parameters are passed in order of the SPL schema.

The function signature must not contain a parameter of the form
``**kwargs`` (`VAR_KEYWORD`).

If there are only formal parameters any non-bound attributes
are not passed into the function.

The SPL schema must have at least the number of positional arguments
the function requires.

Selecting the style
===================

For signatures only containing a parameter of the form 
``*args`` or ``**kwargs`` the style is implicitly defined:

 * ``def f(**tuple_)`` - *dictionary* - ``tuple_`` will contain a dictionary of all of the SPL tuple attribute's values with the keys being the attribute names.
 * ``def f(*tuple_)`` - *tuple* - ``tuple_`` will contain all of the SPL tuple attribute's values in order of the SPL schema definition.

Otherwise the style is set by the ``style`` parameter to the decorator,
defaulting to *attributes by name*. The style value can be set to:

  * ``'name'`` - *attributes by name* (the default)
  * ``'position'`` - *attributes by position*

**Note**: For backwards compatibility ``@spl.pipe`` and ``@spl.sink``
**always** use *attributes by position* and do not support ``**kwargs``.
They do not support the ``style`` parameter.

Examples
========

These examples show how an SPL tuple with the schema and value::

    tuple<rstring id, float64 temp, boolean increase>
    {id='battery', temp=23.7, increase=true}

is passed into a variety of functions by showing the effective Python
call and the resulting values of the function's parameters.

*Dictionary* consuming all attributes by ``**kwargs``::

    @spl.map()
    def f(**tuple_)
        pass
    # f({'id':'battery', 'temp':23.7, 'increase': True})
    #     tuple_={'id':'battery', 'temp':23.7, 'increase':True}

*Tuple* consuming all attributes by ``*args``::

    @spl.map()
    def f(*tuple_)
        pass
    # f('battery', 23.7, True)
    #     tuple_=('battery',23.7, True)

*Attributes by name* consuming all attributes::

    @spl.map()
    def f(id, temp, increase)
        pass
    # f(id='battery', temp=23.7, increase=True)
    #     id='battery'
    #     temp=23.7
    #     increase=True

*Attributes by name* consuming a subset of attributes::

    @spl.map()
    def f(id, temp)
        pass
    # f(id='battery', temp=23.7)
    #    id='battery'
    #    temp=23.7

*Attributes by name* consuming a subset of attributes in a different order::

    @spl.map()
    def f(increase, temp)
        pass
    # f(temp=23.7, increase=True)
    #    increase=True
    #    temp=23.7

*Attributes by name* consuming `id` by name and remaining attributes by ``**kwargs``::

    @spl.map()
    def f(id, **tuple_)
        pass
    # f(id='battery', {'temp':23.7, 'increase':True})
    #    id='battery'
    #    tuple_={'temp':23.7, 'increase':True}

*Attributes by name* consuming `id` by name and remaining attributes by ``*args``::

    @spl.map()
    def f(id, *tuple_)
        pass
    # f(id='battery', 23.7, True)
    #    id='battery'
    #    tuple_=(23.7, True)

*Attributes by position* consuming all attributes::

    @spl.map(style='position')
    def f(key, value, up)
         pass
    # f('battery', 23.7, True)
    #    key='battery'
    #    value=23.7
    #    up=True

*Attributes by position* consuming a subset of attributes::

    @spl.map(style='position')
    def f(a, b)
       pass
    # f('battery', 23.7)
    #    a='battery'
    #    b=23.7

*Attributes by position* consuming `id` by position and remaining attributes by ``*args``::

    @spl.map(style='position')
    def f(key, *tuple_)
        pass
    # f('battery', 23.7, True)
    #    key='battery'
    #    tuple_=(23.7, True)

In all cases the SPL tuple must be able to provide all parameters
required by the function. If the SPL schema is insufficient then
an error will result, typically an SPL compile time error.

The SPL schema can provide a subset of the formal parameters if the
remaining attributes are optional (having a default).

*Attributes by name* consuming a subset of attributes with an optional parameter not matched by the schema::

    @spl.map()
    def f(id, temp, pressure=None)
       pass
    # f(id='battery', temp=23.7)
    #     id='battery'
    #     temp=23.7
    #     pressure=None

.. _submit-from-python:

************************************
Submission of SPL tuples from Python
************************************

The return from a decorated callable results in submission of SPL tuples
on the associated outut port.

A Python function must return:
 * ``None``
 * a Python tuple
 * a Python dictionary
 * a list containing any of the above.

None
====

When ``None`` is return then no tuple will be submitted to
the operator output port.

Python tuple
============

When a Python tuple is returned it is converted to an SPL tuple and
submitted to the output port.

The values of a Python tuple are assigned to an output SPL tuple by position,
so the first value in the Python tuple is assigned to the first attribute
in the SPL tuple::

    # SPL input schema: tuple<int32 x, float64 y>
    # SPL output schema: tuple<int32 x, float64 y, float32 z>
    @spl.map(style='position')
    def myfunc(a,b):
       return (a,b,a+b)

    # The SPL output will be:
    # All values explictly set by returned Python tuple
    # based on the x,y values from the input tuple
    # x is set to: x 
    # y is set to: y
    # z is set to: x+y

The returned tuple may be *sparse*, any attribute value in the tuple
that is ``None`` will be set to their SPL default or copied from
a matching attribute in the input tuple
(same name and type,
or same name and same type as the underlying type of an output attribute
with an optional type),
depending on the operator kind::
    
    # SPL input schema: tuple<int32 x, float64 y>
    # SPL output schema: tuple<int32 x, float64 y, float32 z>
    @spl.map(style='position')
    def myfunc(a,b):
       return (a,None,a+b)

    # The SPL output will be:
    # x is set to: x (explictly set by returned Python tuple)
    # y is set to: y (set by matching input SPL attribute)
    # z is set to: x+y

When a returned tuple has fewer values than attributes in the SPL output
schema the attributes not set by the Python function will be set
to their SPL default or copied from
a matching attribute in the input tuple
(same name and type,
or same name and same type as the underlying type of an output attribute
with an optional type),
depending on the operator kind::
    
    # SPL input schema: tuple<int32 x, float64 y>
    # SPL output schema: tuple<int32 x, float64 y, float32 z>
    @spl.map(style='position')
    def myfunc(a,b):
       return a,

    # The SPL output will be:
    # x is set to: x (explictly set by returned Python tuple)
    # y is set to: y (set by matching input SPL attribute)
    # z is set to: 0 (default int32 value)

When a returned tuple has more values than attributes in the SPL output schema then the additional values are ignored::

    # SPL input schema: tuple<int32 x, float64 y>
    # SPL output schema: tuple<int32 x, float64 y, float32 z>
    @spl.map(style='position')
    def myfunc(a,b):
       return (a,b,a+b,a/b)

    # The SPL output will be:
    # All values explictly set by returned Python tuple
    # based on the x,y values from the input tuple
    # x is set to: x
    # y is set to: y
    # z is set to: x+y
    #
    # The fourth value in the tuple a/b = x/y is ignored.

Python dictionary
=================

A Python dictionary is converted to an SPL tuple for submission to
the associated output port. An SPL attribute is set from the
dictionary if the dictionary contains a key equal to the attribute
name. The value is used to set the attribute, unless the value is
``None``.

If the value in the dictionary is ``None``, or no matching key exists,
then the attribute value is set to its SPL default or copied from
a matching attribute in the input tuple (same name and type,
or same name and same type as the underlying type of an output attribute
with an optional type), depending on the operator kind.

Any keys in the dictionary that do not map to SPL attribute names are ignored.
    
Python list
===========

When a list is returned, each value is converted to an SPL tuple and
submitted to the output port, in order of the list starting with the
first element (position 0). If the list contains `None` at an index
then no SPL tuple is submitted for that index.

The list must only contain Python tuples, dictionaries or `None`. The list
can contain a mix of valid values.

The list may be empty resulting in no tuples being submitted.

"""

from future.builtins import *
from enum import Enum
import functools
import inspect
import re
import sys
import streamsx.ec as ec
import streamsx._streams._runtime
import importlib

import streamsx._streams._version
__version__ = streamsx._streams._version.__version__

############################################
# setup for function inspection
if sys.version_info.major == 3:
  _inspect = inspect
elif sys.version_info.major == 2:
  import funcsigs
  _inspect = funcsigs
else:
  raise ValueError("Python version not supported.")
############################################

# Used to recreate instances of decorated operators
# from their module & class name during pickleling (dill)
# See __reduce__ implementation below
def _recreate_op(op_module, op_name):
    module_ = importlib.import_module(op_module)
    class_ = getattr(module_, op_name)
    return class_.__new__(class_)

_OperatorType = Enum('_OperatorType', 'Ignore Source Sink Pipe Filter Primitive')
_OperatorType.Source.spl_template = 'PythonFunctionSource'
_OperatorType.Pipe.spl_template = 'PythonFunctionPipe'
_OperatorType.Sink.spl_template = 'PythonFunctionSink'
_OperatorType.Filter.spl_template = 'PythonFunctionFilter'
_OperatorType.Primitive.spl_template = 'PythonPrimitive'

_SPL_KEYWORDS = {'graph', 'stream', 'public', 'composite', 'input', 'output', 'type', 'config', 'logic',
                 'window', 'param', 'onTuple', 'onPunct', 'onProcess', 'state', 'stateful', 'mutable',
                 'if', 'for', 'while', 'break', 'continue', 'return', 'attribute', 'function', 'operator'}

def _valid_identifier(id):
    if re.match('^[a-zA-Z_][a-zA-Z_0-9]*$', id) is None or id in _SPL_KEYWORDS:
        raise ValueError("{0} is not a valid SPL identifier".format(id))

def _valid_op_parameter(name):
    _valid_identifier(name)
    if name in ['suppress', 'include']:
        raise ValueError("Parameter name {0} is reserved".format(name))

_EXTRACTING=False

def extracting():
    """Is a module being loaded by ``spl-python-extract``.

    This can be used by modules defining SPL primitive operators
    using decorators such as :py:class:`@spl.map <map>`, to avoid
    runtime behavior. Typically not importing modules that are
    not available locally. The extraction script loads the module
    to determine method signatures and thus does not invoke any methods.

    For example if an SPL toolkit with primitive operators requires
    a package ``extras`` and is using ``opt/python/streams/requirements.txt``
    to include it, then loading it at extraction time can be avoided by::

        from streamsx.spl import spl

        def spl_namespace():
            return 'myns.extras'

        if not spl.extracting():
            import extras

        @spl.map():
        def myextras(*tuple_):
            return extras.process(tuple_)
 
    .. versionadded:: 1.11
    """
    return _EXTRACTING

def pipe(wrapped):
    """
    Decorator to create an SPL operator from a function.
    
    A pipe SPL operator with a single input port and a single
    output port. For each tuple on the input port the
    function is called passing the contents of the tuple.

    SPL attributes from the tuple are passed by position.
    
    The value returned from the function results in
    zero or more tuples being submitted to the operator output
    port, see :ref:`submit-from-python`.

    .. deprecated:: 1.8
        Recommended to use :py:class:`@spl.map <map>` instead.
    """
    if not inspect.isfunction(wrapped):
        raise TypeError('A function is required')

    return _wrapforsplop(_OperatorType.Pipe, wrapped, 'position', False)

#
# Wrap object for an SPL operator, either
# a callable class or function.
#
def _wrapforsplop(optype, wrapped, style, docpy):

    if inspect.isclass(wrapped):
        if not callable(wrapped):
            raise TypeError('Class must be callable')

        _valid_identifier(wrapped.__name__)

        class _op_class(wrapped):
            __doc__ = wrapped.__doc__

            _splpy_wrapped = wrapped
            _splpy_optype = optype
            _splpy_callable = 'class'
            _streamsx_ec_cls = True
            _streamsx_ec_context = streamsx._streams._runtime._has_context_methods(wrapped)

            @functools.wraps(wrapped.__init__)
            def __init__(self,*args,**kwargs):
                super(_op_class, self).__init__(*args,**kwargs)
                self._streamsx_ec_entered = False

            # Use reduce to save the state of the class and its
            # module and operator name.
            def __reduce__(self):
                if hasattr(self, '__getstate__'):
                    state = self.__getstate__()
                else:
                    state = self.__dict__
                return _recreate_op, (wrapped.__module__, wrapped.__name__), state

        if optype in (_OperatorType.Sink, _OperatorType.Pipe, _OperatorType.Filter):
            _op_class._splpy_style = _define_style(wrapped, wrapped.__call__, style)
            _op_class._splpy_fixed_count = _define_fixed(_op_class, _op_class.__call__)
        else:
            _op_class._splpy_style = ''
            _op_class._splpy_fixed_count = -1
     
        _op_class._splpy_file = inspect.getsourcefile(wrapped)
        _op_class._splpy_docpy = docpy
        return _op_class
    if not inspect.isfunction(wrapped):
        raise TypeError('A function or callable class is required')

    _valid_identifier(wrapped.__name__)

    #fnstyle =

    #if fnstyle == 'tuple':
    #    @functools.wraps(wrapped)
    #    def _op_fn(*args):
    #        return wrapped(args)
    #else:
    #    @functools.wraps(wrapped)
    #    def _op_fn(*args, **kwargs):
    #       return wrapped(*args, **kwargs)
    _op_fn = wrapped

    _op_fn._splpy_optype = optype
    _op_fn._splpy_callable = 'function'
    _op_fn._splpy_style = _define_style(_op_fn, _op_fn, style)
    _op_fn._splpy_fixed_count = _define_fixed(_op_fn, _op_fn)
    _op_fn._splpy_file = inspect.getsourcefile(wrapped)
    _op_fn._splpy_docpy = docpy
    _op_fn._streamsx_ec_cls = False
    _op_fn._streamsx_ec_context = False
    return _op_fn

# define the SPL tuple passing style based
# upon the function signature and the decorator
# style parameter
def _define_style(wrapped, fn, style):
    has_args = False
    has_kwargs = False
    has_positional = False
    req_named = False
     
    pmds = _inspect.signature(fn).parameters
    itpmds = iter(pmds)
    # Skip self
    if inspect.isclass(wrapped):
        next(itpmds)

    pc = 0
    for pn in itpmds:
        pmd = pmds[pn]
        if pmd.kind == _inspect.Parameter.POSITIONAL_ONLY:
            raise TypeError('Positional only parameters are not supported:' + pn)
        elif pmd.kind == _inspect.Parameter.VAR_POSITIONAL:
            has_args = True
        elif pmd.kind == _inspect.Parameter.VAR_KEYWORD:
            has_kwargs = True
        elif pmd.kind == _inspect.Parameter.POSITIONAL_OR_KEYWORD:
            has_positional = True
        elif pmd.kind == _inspect.Parameter.KEYWORD_ONLY:
            if pmd.default is _inspect.Parameter.empty:
                req_named = True
        pc +=1
               
    # See if the requested style matches the signature.
    if style == 'position':
        if req_named:
             raise TypeError("style='position' not supported with a required named parameter.")
        elif pc == 1 and has_kwargs:
            raise TypeError("style='position' not supported with single **kwargs parameter.")
        elif pc == 1 and has_args:
            pass
        elif not has_positional:
            raise TypeError("style='position' not supported as no positional parameters exist.")
        # From an implementation point of view the values
        # are passed as a tuple and Python does the correct mapping
        style = 'tuple'

    elif style == 'name':
        if pc == 1 and has_args:
            raise TypeError("style='name' not supported with single *args parameter.")
        elif pc == 1 and has_kwargs:
            raise TypeError("style='name' not supported with single **kwargs parameter.")

    elif style is not None:
        raise TypeError("style=" + style + " unknown.")

    if style is None:
        if pc == 1 and has_kwargs:
            style = 'dictionary'
        elif pc == 1 and has_args:
            style = 'tuple'
        elif pc == 0:
            style = 'tuple'
        else:
            # Default to by name
            style = 'name'

    if style == 'tuple' and has_kwargs:
         raise TypeError("style='position' not implemented with **kwargs parameter.")

    if style == 'name':
         raise TypeError("Not yet implemented!")
    return style

def _define_fixed(wrapped, callable_):
    """For the callable see how many positional parameters are required"""
    is_class = inspect.isclass(wrapped)
    style = callable_._splpy_style if hasattr(callable_, '_splpy_style') else wrapped._splpy_style

    fixed_count = 0
    if style == 'tuple':
        sig = _inspect.signature(callable_)
        pmds = sig.parameters
        itpmds = iter(pmds)
        # Skip 'self' for classes
        if is_class:
            next(itpmds)

        for pn in itpmds:
            param = pmds[pn]
            if param.kind == _inspect.Parameter.POSITIONAL_OR_KEYWORD:
                fixed_count += 1
            if param.kind == _inspect.Parameter.VAR_POSITIONAL: # *args
                fixed_count = -1
                break
            if param.kind == _inspect.Parameter.VAR_KEYWORD:
                break
    return fixed_count

class source(object):
    """
    Create a source SPL operator from an iterable.
    The resulting SPL operator has a single output port.

    When decorating a class the class must be iterable
    having an ``__iter__`` function. When the SPL operator
    is invoked an instance of the class is created
    and an iteration is created using ``iter(instance)``. 

    When decoratiing a function the function must have no
    parameters and must return an iterable or iteration.
    When the SPL operator is invoked the function is called
    and an iteration is created using ``iter(value)``
    where ``value`` is the return of the function.

    For each value in the iteration SPL zero or more tuples
    are submitted to the output port, derived from the value,
    see :ref:`submit-from-python`.
    
    If the iteration completes then no more tuples
    are submitted and a window punctuation mark followed
    by final punctuation mark are submitted to the output port.

    Example definition::

        @spl.source()
        class Range(object):
            def __init__(self, stop, start=0):
                self.start = start
                self.stop = stop

            def __iter__(self):
                return zip(range(self.start, self.stop))

    Example SPL invocation::

        stream<int32 seq> R = Range() {
            param
                stop: 100;
        }

    If ``__iter__`` or ``__next__`` block then shutdown, checkpointing
    or consistent region processing may be delayed. Having ``__next__``
    return ``None`` (no available tuples) or tuples to submit
    will allow such processing to proceed.

    A shutdown ``threading.Event`` is available through
    :py:func:`streamsx.ec.shutdown` which becomes set when a shutdown
    of the processing element has been requested. This event my be waited
    on to perform a sleep that will terminate upon shutdown.
    
    Args:
       docpy: Copy Python docstrings into SPL operator model for SPLDOC.

    Exceptions raised by ``__iter__`` and ``__next__`` can be suppressed
    when this decorator wraps a class with context manager
    ``__enter__`` and ``__exit__`` methods.

    If ``__exit__`` returns a true value when called with an exception 
    then the exception is suppressed.

    Suppressing an exception raised by ``__iter__`` results in the
    source producing an empty iteration. No tuples will be submitted.

    Suppressing an exception raised by ``__next__`` results in the
    source not producing any tuples for that invocation. Processing
    continues with a call to ``__next__``.

    Data conversion errors of the value returned by ``__next__`` can
    also be suppressed by ``__exit__``.
    If ``__exit__`` returns a true value when called with the exception 
    then the exception is suppressed and the value that caused the
    exception is not submitted as an SPL tuple.
    """
    def __init__(self, docpy=True):
        self.style = None
        self.docpy = docpy
    
    def __call__(self, wrapped):
        decorated = _wrapforsplop(_OperatorType.Source, wrapped, self.style, self.docpy)
        if inspect.isclass(decorated):
            decorated._splpy_decor = str(self)
        return decorated

    def __str__(self):
        s = ''
        if not self.docpy:
            if s:
                 s += ', '
            s += 'docpy=False'
        return '@spl.source(' + s + ')'

class map(object):
    """
    Decorator to create a map SPL operator from a callable class or function.

    Creates an SPL operator with a single input port and a single
    output port. For each tuple on the input port the
    callable is called passing the contents of the tuple.

    The value returned from the callable results in
    zero or more tuples being submitted to the operator output
    port, see :ref:`submit-from-python`.

    Example definition::

        @spl.map()
        class AddSeq(object):
        \"\"\"Add a sequence number as the last attribute.\"\"\"
        def __init__(self):
            self.seq = 0

        def __call__(self, *tuple_):
            id = self.seq
            self.seq += 1
            return tuple_ + (id,)

    Example SPL invocation::

        stream<In, tuple<uint64 seq>> InWithSeq = AddSeq(In) { }

    Args:
       style: How the SPL tuple is passed into Python callable or function, see  :ref:`spl-tuple-to-python`.
       docpy: Copy Python docstrings into SPL operator model for SPLDOC.

    Exceptions raised by ``__call__`` can be suppressed when this decorator
    wraps a class with context manager ``__enter__`` and ``__exit__`` methods.
    If ``__exit__`` returns a true value when called with the exception 
    then the exception is suppressed and the tuple that caused the
    exception is dropped.

    Data conversion errors of the value returned by ``__call__`` can
    also be suppressed by ``__exit__``.
    If ``__exit__`` returns a true value when called with the exception 
    then the exception is suppressed and the value that caused the
    exception is not submitted as an SPL tuple.
    """
    def __init__(self, style=None, docpy=True):
        self.style = style
        self.docpy = docpy
    
    def __call__(self, wrapped):
        decorated =  _wrapforsplop(_OperatorType.Pipe, wrapped, self.style, self.docpy)
        if inspect.isclass(decorated):
            decorated._splpy_decor = str(self)
        return decorated

    def __str__(self):
        s = ''
        if self.style is not None:
            s += 'style=' + str(self.style)
        if not self.docpy:
            if s:
                 s += ', '
            s += 'docpy=False'
        return '@spl.map(' + s + ')'

class filter(object):
    """
    Decorator that creates a filter SPL operator from a callable class or function.

    A filter SPL operator has a single input port and one mandatory
    and one optional output port. The schema of each output port
    must match the input port. For each tuple on the input port the
    callable is called passing the contents of the tuple. if the
    function returns a value that evaluates to True then it is
    submitted to mandatory output port 0. Otherwise it it submitted to
    the second optional output port (1) or discarded if the port is
    not specified in the SPL invocation.

    Args:
       style: How the SPL tuple is passed into Python callable or function, see  :ref:`spl-tuple-to-python`.
       docpy: Copy Python docstrings into SPL operator model for SPLDOC.

    Example definition::

        @spl.filter()
        class AttribThreshold(object):
            \"\"\"
            Filter based upon a single attribute being
            above a threshold.
            \"\"\"
            def __init__(self, attr, threshold):
                self.attr = attr
                self.threshold = threshold
                
            def __call__(self, **tuple_):
                return tuple_[self.attr] > self.threshold:

    Example SPL invocation::

        stream<rstring id, float64 voltage> Sensors = ...
        stream<Sensors> InterestingSensors = AttribThreshold(Sensors) {
            param 
              attr: "voltage";
              threshold: 225.0;
        }

    Exceptions raised by ``__call__`` can be suppressed when this decorator
    wraps a class with context manager ``__enter__`` and ``__exit__`` methods.
    If ``__exit__`` returns a true value when called with the exception 
    then the expression is suppressed and the tuple that caused the
    exception is dropped.
    """
    def __init__(self, style=None, docpy=True):
        self.style = style
        self.docpy = docpy
    
    def __call__(self, wrapped):
        decorated =  _wrapforsplop(_OperatorType.Filter, wrapped, self.style, self.docpy)
        if inspect.isclass(decorated):
            decorated._splpy_decor = str(self)
        return decorated

    def __str__(self):
        s = ''
        if self.style is not None:
            s += 'style=' + str(self.style)
        if not self.docpy:
            if s:
                 s += ', '
            s += 'docpy=False'
        return '@spl.filter(' + s + ')'

def ignore(wrapped):
    """
     Decorator to ignore a Python function.

     If a Python callable is decorated with ``@spl.ignore``
     then function is ignored by ``spl-python-extract.py``.

     Args:
         wrapped: Function that will be ignored.
    """
    @functools.wraps(wrapped)
    def _ignore(*args, **kwargs):
        return wrapped(*args, **kwargs)
    _ignore._splpy_optype = _OperatorType.Ignore
    _ignore._splpy_file = inspect.getsourcefile(wrapped)
    return _ignore

# Defines a function as a sink operator
def sink(wrapped):
    """Creates an SPL operator with a single input port.

    A SPL operator with a single input port and no output ports.
    For each tuple on the input port the decorated function
    is called passing the contents of the tuple.

    .. deprecated:: 1.8
        Recommended to use :py:class:`@spl.for_each <for_each>` instead.
    """
    if not inspect.isfunction(wrapped):
        raise TypeError('A function is required')

    return _wrapforsplop(_OperatorType.Sink, wrapped, 'position', False)

# Defines a function as a sink operator
class for_each(object):
    """
    Creates an SPL operator with a single input port.

    A SPL operator with a single input port and no output ports.
    For each tuple on the input port the decorated callable
    is called passing the contents of the tuple.

    Example definition::

        @spl.for_each()
        def PrintTuple(*tuple_):
        \"\"\"Print each tuple to standard out.\"\"\"
            print(tuple_, flush=True)

    Example SPL invocation::

        () as PT = PrintTuple(SensorReadings) { }

    Args:
       style: How the SPL tuple is passed into Python callable, see  :ref:`spl-tuple-to-python`.
       docpy: Copy Python docstrings into SPL operator model for SPLDOC.

    Exceptions raised by ``__call__`` can be suppressed when this decorator
    wraps a class with context manager ``__enter__`` and ``__exit__`` methods.
    If ``__exit__`` returns a true value when called with the exception 
    then the expression is suppressed and the tuple that caused the
    exception is ignored.
    """
    def __init__(self, style=None, docpy=True):
        self.style = style
        self.docpy = docpy

    def __call__(self, wrapped):
        decorated = _wrapforsplop(_OperatorType.Sink, wrapped, self.style, self.docpy)
        if inspect.isclass(decorated):
            decorated._splpy_decor = str(self)
        return decorated

    def __str__(self):
        s = ''
        if self.style is not None:
            s += 'style=' + str(self.style)
        if not self.docpy:
            if s:
                 s += ', '
            s += 'docpy=False'
        return '@spl.for_each(' + s + ')'

class PrimitiveOperator(object):
    """Primitive operator super class.
    Classes decorated with `@spl.primitive_operator` must extend
    this class if they have one or more output ports. This class
    provides the `submit` method to submit tuples to specified
    otuput port.

    .. versionadded:: 1.8
    """
    def submit(self, port_id, tuple_):
        """Submit a tuple to the output port.

        The value to be submitted (``tuple_``) can be a ``None`` (nothing will be submitted),
        ``tuple``, ``dict` or ``list`` of those types. For details
        on how the ``tuple_`` is mapped to an SPL tuple see :ref:`submit-from-python`.

        Args:
             port_id: Identifier of the port specified in the
                  ``output_ports`` parameter of the ``@spl.primitive_operator``
                  decorator.
             tuple_: Tuple (or tuples) to be submitted to the output port.
        """
        port_index = self._splpy_output_ports[port_id]
        ec._submit(self, port_index, tuple_)

    def all_ports_ready(self):
        """Notifcation that the operator can submit tuples.

        Called when the primitive operator can submit tuples
        using :py:meth:`submit`. An operator must not submit
        tuples until this method is called or until a port
        processing method is called.

        Any implementation must not block. A typical use
        is to start threads that submit tuples.

        An implementation must return a value that allows
        the SPL runtime to determine when an operator completes.
        An operator completes, and finalizes its output ports
        when:

            * All input ports (if any) have been finalized.
            * All background processing is complete.

        The return from ``all_ports_ready`` defines when
        background processing, such as threads started by
        ``all_ports_ready``, is complete. The value is one of:

            * A value that evaluates to `False` - No background processing exists.
            * A value that evaluates to `True` - Background processing exists and never completes. E.g. a source operator that processes real time events.
            * A callable - Background processing is complete when the callable returns. The SPL runtime invokes the callable once (passing no arguments) when the method returns background processing is assumed to be complete.

        For example if an implementation starts a single thread then `Thread.join` is returned to complete the operator when the thread completes::

            def all_ports_ready(self):
                submitter = threading.Thread(target=self._find_and_submit_data)
                submitter.start()
                return submitter.join

            def _find_and_submit_data(self):
                ...

        Returns:
            Value indicating active background processing.
        

        This method implementation does nothing and returns ``None``.
        """
        return None


class input_port(object):
    """Declare an input port and its processor method.

    Instance methods within a class decorated by
    :py:class:`spl.primitive_operator <primitive_operator>` declare
    input ports by decorating methods with this decorator.

    Each tuple arriving on the input port will result in a call
    to the processor method passing the stream tuple converted to
    a Python representation depending on the style. The style is
    determined by the method signature or the `style` parameter,
    see  :ref:`spl-tuple-to-python`.

    The order of the methods within the class define
    the order of the ports, so the first port is
    the first method decorated with `input_port`.

    Args:
        style: How the SPL tuple is passed into the method, see  :ref:`spl-tuple-to-python`.

    .. versionadded:: 1.8
    """
    _count = 0
    def __init__(self, style=None):
        self._style = style

    def __call__(self, wrapped):
        wrapped._splpy_input_port_seq = input_port._count
        wrapped._splpy_input_port_config = self
        wrapped._splpy_style = self._style
        input_port._count += 1
        return wrapped


class primitive_operator(object):
    """Creates an SPL primitive operator with an arbitrary number of input ports and
    output ports.

    Input ports are declared by decorating an instance method
    with :py:meth:`input_port`. The method is the process method
    for the input port and is called for each tuple that arrives
    at the port. The order of the decorated process methods defines
    the order of the ports in the SPL operator, with the first
    process method being the first port at index zero.

    Output ports are declared by the ``output_ports`` parameter which
    is set to a ``list`` of port identifiers. The port identifiers are
    arbitrary but must be hashable. Port identifiers allow the ability
    to submit tuples "logically' rather than through a port index. Typically
    a port identifier will be a `str` or an `enum`. The size of the list
    defines the number of output ports with the first identifier in the list
    coresponding to the first output port of the operator at index zero.
    If the list is empty or not set then the operator has no output ports.

    Tuples are submitted to an output port using :py:meth:`~PrimitiveOperator.submit`.

    When an operator has output ports it must be a sub-class of
    :py:class:`PrimitiveOperator` which provides the
    :py:meth:`~PrimitiveOperator.submit` method and the ports
    ready notification mechanism :py:meth:`~PrimitiveOperator.all_ports_ready`.

    Example definition of an operator with a single input port and two output ports::

        @spl.primitive_operator(output_ports=['MATCH', 'NEAR_MATCH'])
        class SelectCustomers(spl.PrimitiveOperator):
            \"\"\" Score customers using a model.
            Customers that are a good match are submitted to port 0 ('MATCH')
            while customers that are a near match are submitted to port 1 ('NEAR_MATCH').

            Customers that are not a good or near match are not submitted to any port.
            \"\"\"
            def __init__(self, match, near_match):
                self.match = match
                self.near_match = near_match

            @spl.input_port()
            def customers(self, **tuple_):
                 customer_score = self.score(tuple_)
                 if customer_score >= self.match:
                     self.submit('MATCH', tuple_)
                 elif customer_score >= self.near_match:
                     self.submit('NEAR_MATCH', tuple_)

            def score(self, **customer):
                # Actual model scoring omitted
                score = ...
                return score

    Example SPL invocation::

        (stream<Customers> MakeOffer; stream<Customers> ImproveOffer>) = SelectCustomers(Customers) {
            param
                match: 0.9;
                near_match: 0.8;
        }


    Args:
       output_ports(list): List of identifiers for output ports.
       docpy: Copy Python docstrings into SPL operator model for SPLDOC.

    .. versionadded:: 1.8
    """

    def __init__(self, output_ports=None,docpy=True):
        self._docpy = docpy
        self._output_ports = output_ports

    def __call__(self, wrapped):
        if not inspect.isclass(wrapped):
            raise TypeError('A class is required:' + str(wrapped))

        _valid_identifier(wrapped.__name__)

        cls = _wrapforsplop(_OperatorType.Primitive, wrapped, None, self._docpy)

        inputs = dict()
        for fname, fn in inspect.getmembers(wrapped):
            if hasattr(fn, '_splpy_input_port_seq'):
                if sys.version_info.major == 2:
                    fn = fn.__func__
                inputs[fn._splpy_input_port_seq] = fn

        cls._splpy_input_ports = []
        cls._splpy_style = []
        cls._splpy_fixed_count = []
        for seq in sorted(inputs.keys()):
            fn = inputs[seq]
            fn._splpy_input_port_id = len(cls._splpy_input_ports)
            fn._splpy_style = _define_style(wrapped, fn, fn._splpy_style)
            fn._splpy_fixed_count = _define_fixed(cls, fn)

            cls._splpy_input_ports.append(fn)
            cls._splpy_style.append(fn._splpy_style)
            cls._splpy_fixed_count.append(fn._splpy_fixed_count)

        cls._splpy_output_ports = dict()
        if self._output_ports:
            for i in range(len(self._output_ports)):
                cls._splpy_output_ports[self._output_ports[i]] = i

        cls._splpy_decor = str(self)
        return cls

    def __str__(self):
        s = ''
        if self._output_ports:
            s += 'output_ports=' + str(self._output_ports)
             
        if not self._docpy:
            if s:
                 s += ', '
            s += 'docpy=False'
        return '@spl.primitive(' + s + ')'



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2018

## Common PEP396 version for modules shipped with streamsx packages
__version__='1.11.8'

import pkg_resources
import sys
import warnings


# Attempt to detect cases where the python code does not
# match the package install. Typically due to having
# PYTHONPATH point into an older topology toolkit as
# is set by default by streamsprofile.sh
def _mismatch_check(module_name):
    srp = pkg_resources.working_set.find(pkg_resources.Requirement.parse('streamsx'))
    if srp is None or not srp.has_version():
        spv = pkg_resources.parse_version(__version__)
    else:
        spv = srp.parsed_version

    if module_name not in sys.modules:
        return

    sm = sys.modules[module_name]

    file_name = sm.__file__ if hasattr(sm, '__file__') else 'unknown location'

    # Pre 1.11 versions did not maintain a version
    # number so they must be older than this code
    # added in 1.11
    if not hasattr(sm, '__version__'):
        warnings.warn(_warning_msg(spv, module_name, file_name))
        return

    opv = pkg_resources.parse_version(sm.__version__)
    opv_base = pkg_resources.parse_version(opv.base_version)
    spv_base = pkg_resources.parse_version(spv.base_version)
    if opv_base < spv_base:
        warnings.warn(_warning_msg(spv, module_name, file_name, opv))
        return

def _warning_msg(spv, module_name, file_name, version=None):
    version = ": " + str(version) if version else ""
    return "Version mismatch: streamsx " + str(spv) + " is installed but " + module_name + " from " + file_name +  " is at an older version" + version + "."



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2018

from __future__ import print_function
from future.builtins import *

import sys
import os
import argparse
import pkg_resources
import streamsx._streams._version
import streamsx.topology.context

def main(args=None):
    """ Output information about `streamsx` and the environment.
   
        Useful for support to get key information for use of `streamsx`
        and Python in IBM Streams.
    """
    _parse_args(args)
    streamsx._streams._version._mismatch_check('streamsx.topology.context')
    srp = pkg_resources.working_set.find(pkg_resources.Requirement.parse('streamsx'))
    if srp is not None:
        srv = srp.parsed_version
        location = srp.location
        spkg = 'package'
    else:
        srv = streamsx._streams._version.__version__
        location = os.path.dirname(streamsx._streams._version.__file__)
        location = os.path.dirname(location)
        location = os.path.dirname(location)
        tk_path = (os.path.join('com.ibm.streamsx.topology', 'opt', 'python', 'packages'))
        spkg = 'toolkit' if location.endswith(tk_path) else 'unknown'

    print('streamsx==' + str(srv) + ' (' + spkg + ')')
    print('  location: ' + str(location))
    print('Python version:' + str(sys.version))
    print('PYTHONHOME=' + str(os.environ.get('PYTHONHOME', 'unset')))
    print('PYTHONPATH=' + str(os.environ.get('PYTHONPATH', 'unset')))
    print('PYTHONWARNINGS=' + str(os.environ.get('PYTHONWARNINGS', 'unset')))
    print('STREAMS_INSTALL=' + str(os.environ.get('STREAMS_INSTALL', 'unset')))
    print('JAVA_HOME=' + str(os.environ.get('JAVA_HOME', 'unset')))
    return 0

def _parse_args(args):
    """ Argument parsing
    """
    cmd_parser = argparse.ArgumentParser(description='Prints support information about streamsx package and environment.')
    return cmd_parser.parse_args(args)

if __name__ == '__main__':
    sys.exit(main())



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2015,2017

from __future__ import print_function
from future.builtins import *

import sys

# Print function that flushes
def print_flush(v):
    """
    Prints argument to stdout flushing after each tuple.
    :returns: None
    """
    print(v)
    sys.stdout.flush()

def identity(t):
    """
    Returns its single argument.
    :returns: Its argument.
    """
    return t;




# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2016,2018
from __future__ import print_function
from __future__ import unicode_literals
from future.builtins import *
import sys
import sysconfig
import inspect
if sys.version_info.major == 2:
  import funcsigs
import imp
import glob
import fcntl
import fnmatch
import os
import shutil
import argparse
import subprocess
import xml.etree.ElementTree as ET
import html
from streamsx.spl.spl import _OperatorType
from streamsx.spl.spl import _valid_op_parameter
import streamsx.spl.spl

import streamsx._streams._version
__version__ = streamsx._streams._version.__version__

############################################
# setup for function inspection
if sys.version_info.major == 3:
    _inspect = inspect
elif sys.version_info.major == 2:
    _inspect = funcsigs
    FileNotFoundError = IOError
else:
    raise ValueError("Python version not supported.")
############################################

# Return the root of the com.ibm.streamsx.topology toolkit
def _topology_tk_dir():
    dir_ = os.path.dirname(__file__) # streamsx/scripts
    dir_ = os.path.dirname(dir_) # streamsx
 
    # See if we are being run from streamsx package
    pkg_tk = os.path.join(dir_, '.toolkit', 'com.ibm.streamsx.topology')
    if os.path.isdir(pkg_tk):
        return pkg_tk

    # Run from the SPL toolkit
    for _ in range(4):
        dir_ = os.path.dirname(dir_)
    return dir_

def replaceTokenInFile(file, token, value):
    f = open(file,'r')
    contents = f.read()
    f.close()

    newcontents = contents.replace(token, value)

    f = open(file,'w')
    f.write(newcontents)
    f.close()

def _optype(opobj):
    if hasattr(opobj, '_splpy_optype'):
        return opobj._splpy_optype
    return None

def _opfile(opobj):
    return opobj._splpy_file

def _opcallable(opobj):
    return opobj._splpy_callable

def _opdoc(opobj):
    return opobj._splpy_docpy

_INFO_XML_TEMPLATE="""<?xml version="1.0" encoding="UTF-8"?>
<toolkitInfoModel
  xmlns="http://www.ibm.com/xmlns/prod/streams/spl/toolkitInfo"
  xmlns:common="http://www.ibm.com/xmlns/prod/streams/spl/common">
  <identity>
    <name>__SPLPY_TOOLKIT_NAME__</name>
    <description>Automatic generated toolkit description file.</description>
    <version>1.0.0</version>
    <requiredProductVersion>4.2</requiredProductVersion>
  </identity>
  <dependencies/>
  <resources>
    <messageSet name="TopologySplpyResource">
      <lang default="true">en_US/TopologySplpyResource.xlf</lang>
      <lang>de_DE/TopologySplpyResource.xlf</lang>
      <lang>es_ES/TopologySplpyResource.xlf</lang>
      <lang>fr_FR/TopologySplpyResource.xlf</lang>
      <lang>it_IT/TopologySplpyResource.xlf</lang>
      <lang>ja_JP/TopologySplpyResource.xlf</lang>
      <lang>ko_KR/TopologySplpyResource.xlf</lang>
      <lang>pt_BR/TopologySplpyResource.xlf</lang>
      <lang>ru_RU/TopologySplpyResource.xlf</lang>
      <lang>zh_CN/TopologySplpyResource.xlf</lang>
      <lang>zh_TW/TopologySplpyResource.xlf</lang>
    </messageSet>
  </resources>
</toolkitInfoModel>
"""


# Create SPL operator parameters from the Python class
# (functions cannot have parameters)
# The parameters are taken from the signature of
# the __init__ method. In the spirit of Python
# the default for non-annotated function parameters
# is to map to operator parameters that take any type
# with a cardinality of 1. If the function parameter
# has a default value, then the operator parameter is optional

_OP_PARAM_TEMPLATE ="""
 <parameter>
  <name>__SPLPY__PARAM_NAME__SPLPY__</name>
  <description></description>
  <optional>__SPLPY__PARAM_OPT__SPLPY__</optional>
  <rewriteAllowed>true</rewriteAllowed>
  <expressionMode>AttributeFree</expressionMode>
  <type></type>
  <cardinality>1</cardinality>
 </parameter>"""

_OP_INPUT_PORT_SET_TEMPLATE ="""
<inputPortSet>
  <description>
                                    __SPLPY__INPORT_DESCRIPTION__SPLPY__
  </description>
  <tupleMutationAllowed>false</tupleMutationAllowed>
  <windowingMode>NonWindowed</windowingMode>
  <windowPunctuationInputMode>Oblivious</windowPunctuationInputMode>
  <cardinality>1</cardinality>
  <optional>false</optional>
</inputPortSet>
"""

_OP_OUTPUT_PORT_SET_TEMPLATE ="""
<outputPortSet>
  <description>
    __SPLPY__OUTPORT_DESCRIPTION__SPLPY__
  </description>
  <expressionMode>Nonexistent</expressionMode>
  <autoAssignment>false</autoAssignment>
  <completeAssignment>false</completeAssignment>
  <rewriteAllowed>true</rewriteAllowed>
  <windowPunctuationOutputMode>Free</windowPunctuationOutputMode>
  <tupleMutationAllowed>false</tupleMutationAllowed>
  <cardinality>1</cardinality>
  <optional>false</optional>
</outputPortSet>
"""





class _Extractor(object):
    def __init__(self, args):
        self._cmd_args = self._parse_cmd_args(args)
        self._tk_dir = self._cmd_args.directory

    def _parse_cmd_args(self, args):
        cmd_parser = argparse.ArgumentParser(description='Extract SPL operators from decorated Python classes and functions.')
        cmd_parser.add_argument('-i', '--directory', required=True,
                   help='Toolkit directory')
        cmd_parser.add_argument('--make-toolkit', action='store_true',
                   help='Index toolkit using spl-make-toolkit')
        cmd_parser.add_argument('-v', '--verbose', action='store_true',
                   help='Print more diagnostics')
        return cmd_parser.parse_args(args)

    def _make_namespace_dir(self, ns):
         nsdir = os.path.join(self._tk_dir, ns)
         if os.path.isdir(nsdir):
             return nsdir
         os.mkdir(nsdir)
         return nsdir

    def _make_operator_dir(self, nsdir, name):
         oppath = os.path.join(nsdir, name)
         if (os.path.isdir(oppath)):
             shutil.rmtree(oppath)
         os.mkdir(oppath)
         return oppath

    def _make_toolkit(self):
        # Now make the toolkit if required
        if self._cmd_args.make_toolkit:
            si = os.environ['STREAMS_INSTALL']
            mktk = os.path.join(si, 'bin', 'spl-make-toolkit')
            mktk_args = [mktk, '--directory', self._cmd_args.directory, '--make-operator']
            subprocess.check_call(mktk_args)

    # Process python objects in a module looking for SPL operators
    # dynm - introspection for the modeul
    # module - module name
    # ops - list of potential operators (functions)
    def _process_operators(self, dynm, module, streams_python_file, ops):
        for opname, opobj in ops:
            if inspect.isbuiltin(opobj):
                continue
            if opname.startswith('spl'):
                continue
            optype = _optype(opobj)
            if optype is None:
                continue
            if optype == _OperatorType.Ignore:
                continue
            if streams_python_file != _opfile(opobj):
                continue
            self._common_tuple_operator(dynm, module, opname, opobj)

    def _copy_globalization_resources(self):
        '''Copy the language resource files for python api functions
    
        This function copies the TopologySplpy Resource files from Topology toolkit directory
        into the impl/nl folder of the project.
        Returns: the list with the copied locale strings'''
        rootDir = os.path.join(_topology_tk_dir(), "impl", "nl")
        languageList = []
        for dirName in os.listdir(rootDir):
            srcDir = os.path.join(_topology_tk_dir(), "impl", "nl", dirName)
            if (os.path.isdir(srcDir)) and (dirName != "include"):
                dstDir = os.path.join(self._tk_dir, "impl", "nl", dirName)
                try:
                    print("Copy globalization resources " + dirName)
                    os.makedirs(dstDir)
                except OSError as e:
                    if (e.errno == 17) and (os.path.isdir(dstDir)):
                        if self._cmd_args.verbose:
                            print("Directory", dstDir, "exists")
                    else:
                        raise
                srcFile = os.path.join(srcDir, "TopologySplpyResource.xlf")
                if os.path.isfile(srcFile):
                    res = shutil.copy2(srcFile, dstDir)
                    languageList.append(dirName)
                    if self._cmd_args.verbose:
                        print("Written: " + res)
        return languageList

    #
    # module - module for operator
    # opname - name of the SPL operator
    # opobj - decorated object defining operator
    #
    def _common_tuple_operator(self, dynm, module, opname, opobj) :        
        if (not hasattr(dynm, 'spl_namespace')) and hasattr(dynm, 'splNamespace'):
            ns = getattr(dynm, 'splNamespace')()
        else:
            ns = getattr(dynm, 'spl_namespace')()
        print(ns + "::" + opname)
        # Print the summary of the class/function
        _doc = inspect.getdoc(opobj)
        if _doc is not None:
            _doc = _doc.splitlines()[0]
            print("  ", _doc)
        nsdir = self._make_namespace_dir(ns)
        opdir = self._make_operator_dir(nsdir, opname)
        self._copy_template_dir("common")
        self._copy_template_dir("icons")
        self._copy_python_dir("packages")
        self._copy_python_dir("include")
        self._copy_CGT(opdir, ns, opname, opobj)
        self._write_config(dynm, opdir, module, opname, opobj)

    def _create_op_parameters(self, opmodel_xml, name, opObj):
        opparam_xml = ''
        if _opcallable(opObj) == 'class':
            pmds = init_sig = _inspect.signature(opObj._splpy_wrapped.__init__).parameters
            itpmds = iter(pmds)
            # first argument to __init__ is self (instance ref)
            next(itpmds)
        
            for pn in itpmds:
                pmd = pmds[pn]
                _valid_op_parameter(pn)
                px = _OP_PARAM_TEMPLATE
                px = px.replace('__SPLPY__PARAM_NAME__SPLPY__', pn)
                px = px.replace('__SPLPY__PARAM_OPT__SPLPY__', 'false' if pmd.default== _inspect.Parameter.empty else 'true' )
                opparam_xml = opparam_xml + px
        replaceTokenInFile(opmodel_xml, '__SPLPY__PARAMETERS__SPLPY__', opparam_xml)
    
    def _copy_CGT(self, opdir, ns, name, funcTuple):
         cgtbase = _optype(funcTuple).spl_template
         optemplate = os.path.join(_topology_tk_dir(), "opt", "python", "templates","operators", cgtbase)
         opcgt_cpp = os.path.join(opdir, name + '_cpp.cgt')
         shutil.copy(optemplate + '_cpp.cgt', opcgt_cpp)
         shutil.copy(optemplate + '_h.cgt', os.path.join(opdir, name + '_h.cgt'))
         opmodel_xml = os.path.join(opdir, name + '.xml')
         shutil.copy(optemplate + '.xml', opmodel_xml)
         replaceTokenInFile(opmodel_xml, "__SPLPY__MAJOR_VERSION__SPLPY__", str(sys.version_info.major));
         replaceTokenInFile(opmodel_xml, "__SPLPY__MINOR_VERSION__SPLPY__", str(sys.version_info.minor));
         self._create_op_parameters(opmodel_xml, name, funcTuple)
         self._create_op_spldoc(opmodel_xml, name, funcTuple)
         if cgtbase == 'PythonPrimitive':
             self._create_primitive(opmodel_xml, name, funcTuple)
         else:
             self._create_ip_spldoc(opmodel_xml, name, funcTuple)

    ## Create SPL doc entries in the Operator model xml file.
    ##
    def _create_op_spldoc(self, opmodel_xml, name, opobj):
         opdoc = inspect.getdoc(opobj)
         if opdoc is None:
             opdoc = 'Callable: ' + name + "\n"

         opdoc = html.escape(opdoc)

         # Optionally include the Python source code
         if _opdoc(opobj):
             decor = None
             if inspect.isclass(opobj):
                 if hasattr(opobj, '_splpy_decor'):
                      decor = opobj._splpy_decor
                 opobj = inspect.getmro(opobj)[1]
             try:
                 _pysrc = inspect.getsource(opobj)
                 opdoc += "\n"
                 opdoc += "# Python\n";

                 if decor:
                     opdoc += '    '
                     opdoc += decor
                     opdoc += '\n'

                 for _line in str.splitlines(_pysrc):
                     opdoc += '    '
                     opdoc += html.escape(_line)
                     opdoc += '\n'
             except:
                 pass
         
         replaceTokenInFile(opmodel_xml, "__SPLPY__DESCRIPTION__SPLPY__", opdoc);

    def _create_ip_spldoc(self, opmodel_xml, name, opobj):
         if opobj._splpy_style == 'dictionary':
           _p0doc = """
           Tuple attribute values are passed by name to the Python callable using `\*\*kwargs`.
                 """
         elif opobj._splpy_style == 'tuple':
           _p0doc = """
           Tuple attribute values are passed by position to the Python callable.
                 """
         else:
           _p0doc = ''
     
         replaceTokenInFile(opmodel_xml, "__SPLPY__INPORT_0_DESCRIPTION__SPLPY__", _p0doc);

    def _create_primitive(self, opmodel_xml, name, cls):
        # input ports
        if cls._splpy_input_ports:
            ipd = ''
            for portfn in cls._splpy_input_ports:
                tip = _OP_INPUT_PORT_SET_TEMPLATE
                fndoc = inspect.getdoc(portfn)
                if not fndoc:
                    fndoc = portfn.__name__
                tip = tip.replace('__SPLPY__INPORT_DESCRIPTION__SPLPY__', fndoc)
                ipd += tip
        else:
            ipd = "<!-- no input ports -->"
        replaceTokenInFile(opmodel_xml, "__SPLPY__INPORTS__SPLPY__", ipd);

        # output ports
        if cls._splpy_output_ports:
            opd = ''
            for portfn in cls._splpy_output_ports:
                top = _OP_OUTPUT_PORT_SET_TEMPLATE
                opd += top
        else:
            opd = "<!-- no output ports -->"
        replaceTokenInFile(opmodel_xml, "__SPLPY__OUTPORTS__SPLPY__", opd);
   
    # Write information about the Python function parameters.
    #
    def _write_style_info(self, cfgfile, opobj):
            style = opobj._splpy_style

            if isinstance(style, list):
                pm_style = '(' + ','.join(["'{0}'".format(_) for _ in opobj._splpy_style]) + ')'
            else:
                pm_style = "'" + style + "'"
            cfgfile.write('sub splpy_ParamStyle {'+ pm_style + '}\n')
             
            if isinstance(opobj._splpy_fixed_count, list):
                pm_fixed = '(' + ','.join(["'{0}'".format(_) for _ in opobj._splpy_fixed_count]) + ')'
            else:
                pm_fixed = "'" + str(opobj._splpy_fixed_count) + "'"

            cfgfile.write('sub splpy_FixedParam { '+ pm_fixed + "}\n")
 

    # Write out the configuration for the operator
    # as a set of Perl functions that return useful values
    # for the code generator
    def _write_config(self, dynm, opdir, module, opname, opobj):
        cfgpath = os.path.join(opdir, 'splpy_operator.pm')
        cfgfile = open(cfgpath, 'w')
        cfgfile.write('sub splpy_Module { \''+ module   + "\'}\n")
        cfgfile.write('sub splpy_OperatorCallable {\'' + _opcallable(opobj) + "\'}\n")
        cfgfile.write('sub splpy_FunctionName {\'' + opname + "\'}\n")
        cfgfile.write('sub splpy_OperatorType {\'' + _optype(opobj).name + "\'}\n")

        cfgfile.write('sub splpy_PyMajor {' + str(sys.version_info.major) + "}\n")
        cfgfile.write('sub splpy_PyMinor {' + str(sys.version_info.minor) + "}\n")
        self._write_style_info(cfgfile, opobj)

        if hasattr(dynm, 'spl_pip_packages'):
            pp = getattr(dynm, 'spl_pip_packages')()
            if not isinstance(pp, list):
                pp = list(pp)
        else:
            pp = []

        cfgfile.write('sub splpy_Packages {(' + ','.join(["'{0}'".format(_) for _ in pp]) + ')}\n')

        cfgfile.write("1;\n")
        cfgfile.close()

    # Copy a single file from the templates directory to the newly created operator directory
    def _copy_template_dir(self, dir_):
        self._copy_python_dir(os.path.join("templates", dir_))

    def _copy_python_dir(self, dir_):
        cmn_src = os.path.join(_topology_tk_dir(), "opt", "python", dir_);
        cmn_dst = os.path.join(self._tk_dir, "opt", ".splpy", os.path.basename(dir_))
        if os.path.isdir(cmn_dst):
            shutil.rmtree(cmn_dst)
        # Remove the old directory used - see #1997
        old_dst = os.path.join(self._tk_dir, "opt", ".__splpy")
        if os.path.isdir(old_dst):
            shutil.rmtree(old_dst)
        shutil.copytree(cmn_src, cmn_dst)

    def _setup_info_xml(self, languageList):
        '''Setup the info.xml file
        
        This function prepares or checks the info.xml file in the project directory
        - if the info.xml does not exist in the project directory, it copies the template info.xml into the project directory.
          The project name is obtained from the project directory name
        - If there is a info.xml file, the resource section is inspected. If the resource section has no valid message set
          description for the TopologySplpy Resource a warning message is printed'''
        infoXmlFile = os.path.join(self._tk_dir, 'info.xml')
        print('Check info.xml:', infoXmlFile)
        try:
            TopologySplpyResourceMessageSetFound = False
            TopologySplpyResourceLanguages = []
            tree = ET.parse(infoXmlFile)
            root = tree.getroot()
            for resources in root.findall('{http://www.ibm.com/xmlns/prod/streams/spl/toolkitInfo}resources'):
                if self._cmd_args.verbose:
                    print('Resource: ', resources.tag)
                for messageSet in resources.findall('{http://www.ibm.com/xmlns/prod/streams/spl/toolkitInfo}messageSet'):
                    if self._cmd_args.verbose:
                        print('Message set:', messageSet.tag, messageSet.attrib)
                    if 'name' in messageSet.attrib:
                        if messageSet.attrib['name'] == 'TopologySplpyResource':
                            TopologySplpyResourceMessageSetFound = True
                            for lang in messageSet.findall('{http://www.ibm.com/xmlns/prod/streams/spl/toolkitInfo}lang'):
                                language = os.path.dirname(lang.text)
                                TopologySplpyResourceLanguages.append(language)
            if TopologySplpyResourceMessageSetFound:
                TopologySplpyResourceLanguages.sort()
                languageList.sort()
                copiedLanguagesSet = set(languageList)
                resourceLanguageSet = set(TopologySplpyResourceLanguages)
                if self._cmd_args.verbose:
                    print('copied language resources:\n', languageList)
                    print('TopologySplpyResource from info.xml:\n', TopologySplpyResourceLanguages)
                if copiedLanguagesSet == resourceLanguageSet:
                    print('Resource section of info.xml verified')
                else:
                    errstr = """"ERROR: Message set for the "TopologySplpyResource" is incomplete or invalid. Correct the resource section in info.xml file.
    
                    Sample info xml:\n""" + _INFO_XML_TEMPLATE
                    sys.exit(errstr)
            else:
                errstr = """"ERROR: Message set for the "TopologySplpyResource" is missing. Correct the resource section in info.xml file.
    
                    Sample info xml:\n""" + _INFO_XML_TEMPLATE
                sys.exit(errstr)
        except FileNotFoundError as e:
            print("WARNING: File info.xml not found. Creating info.xml from template")
            #Get default project name from project directory
            projectRootDir = os.path.abspath(self._tk_dir) #os.path.abspath returns the path without trailing /
            projectName = os.path.basename(projectRootDir)
            infoXml=_INFO_XML_TEMPLATE.replace('__SPLPY_TOOLKIT_NAME__', projectName)
            f = open(infoXmlFile, 'w')
            f.write(infoXml)
            f.close()
        except SystemExit as e:
            raise e
        except:
            errstr = """ERROR: File info.xml is invalid or not accessible
    
                Sample info xml:\n""" + _INFO_XML_TEMPLATE
            sys.exit(errstr)

def _extract_from_toolkit(args):
    """
    Look at all the modules in opt/python/streams (opt/python/streams/*.py)
    and extract any spl decorated function as an operator.
    """

    extractor = _Extractor(args)
    if extractor._cmd_args.verbose:
        print("spl-python-extract:", __version__)
        print("Topology toolkit location:", _topology_tk_dir())

    tk_dir = extractor._tk_dir

    tk_streams = os.path.join(tk_dir, 'opt', 'python', 'streams')
    if not os.path.isdir(tk_streams) or not fnmatch.filter(os.listdir(tk_streams), '*.py'):
        # Nothing to do for Python extraction
        extractor._make_toolkit()
        return

    lf = os.path.join(tk_streams, '.lockfile')
    with  open(lf, 'w') as lfno:
        fcntl.flock(lfno, fcntl.LOCK_EX)

        tk_idx = os.path.join(tk_dir, 'toolkit.xml')
        tk_time = os.path.getmtime(tk_idx) if os.path.exists(tk_idx) else None
        changed = False if tk_time else True
        if tk_time:
            for mf in glob.glob(os.path.join(tk_streams, '*.py')):
               if os.path.getmtime(mf) >= tk_time:
                   changed = True
                   break

        if changed:
            path_items = _setup_path(tk_dir, tk_streams)

            for mf in glob.glob(os.path.join(tk_streams, '*.py')):
                print('Checking ', mf, 'for operators')
                name  = inspect.getmodulename(mf)
                dynm = imp.load_source(name, mf)
                streams_python_file = inspect.getsourcefile(dynm)
                extractor._process_operators(dynm, name, streams_python_file, inspect.getmembers(dynm, inspect.isfunction))
                extractor._process_operators(dynm, name, streams_python_file, inspect.getmembers(dynm, inspect.isclass))

            langList = extractor._copy_globalization_resources()
            if extractor._cmd_args.verbose:
                print("Available languages for TopologySplpy resource:", langList)
            extractor._setup_info_xml(langList)

            extractor._make_toolkit()

            _reset_path(path_items)
            fcntl.flock(lfno, fcntl.LOCK_UN)

def _setup_path(tk_dir, tk_streams):
    sys.path.insert(1, tk_streams)
    items = [tk_streams]
    tk_packages = os.path.join(tk_dir, 'opt', 'python', 'packages')
    if os.path.isdir(tk_packages):
        sys.path.insert(1, tk_packages)
        items.append(tk_packages)
    tk_modules = os.path.join(tk_dir, 'opt', 'python', 'modules')
    if os.path.isdir(tk_modules):
        sys.path.insert(1, tk_modules)
        items.append(tk_modules)
    return items

def _reset_path(items):
    for p in items:
        sys.path.remove(p)

def main(args=None):
    streamsx._streams._version._mismatch_check('streamsx.spl.spl')
    try:
        streamsx.spl.spl._EXTRACTING = True
        _extract_from_toolkit(args)
    finally:
        streamsx.spl.spl._EXTRACTING = False

if __name__ == '__main__':
    main()



import nose.plugins.base
import os



class JCO(nose.plugins.base.Plugin):
    enabled = False
    name = 'streamsx-jco'
    score = 1000

    def options(self, parser, env=os.environ):
        print('OPTIONS', flush=True)
        super(JCO, self).options(parser, env=env)

    def configure(self, options, conf):
        super(JCO, self).configure(options, conf)
        JCO.enabled = options.enable_plugin_streamsx_jco

    def beforeTest(self, test):
        print('BEFORE', test, type(test.test), flush=True)
        print('BEFORE', test.test, flush=True)
        #print('BEFORE', test.test.test_ctxtype, flush=True)
        test.test.test_jco = {'fred': 3}



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2016

"""
Publish and subscribe to MQTT messages.

Additional information at http://mqtt.org and
http://ibmstreams.github.io/streamsx.messaging
"""

from future.builtins import *

from streamsx.topology.topology import *
from streamsx.topology import schema

class MqttStreams(object):
    """
    A simple connector to a MQTT broker for publishing
    string tuples to MQTT topics, and
    subscribing to MQTT topics and creating streams.
    
    A connector is for a specific MQTT Broker as specified in
    the configuration object config. Any number of  publish()and  subscribe()
    connections may be created from a single mqtt_streams connector.
     
    Sample use:
    ::
        topo = Topology("An MQTT application")
        # define configuration information
        config = {}
        config['clientID'] = "test_MQTTpublishClient"
        config['qos'] = int("1") #(needs to be int vs long)
        config['keepAliveInterval'] = int(20) (needs to be int vs long)
        config['commandTimeout'] = 30000 (needs to be int vs long)
        config['period'] = 5000 (needs to be int vs long)
        config['messageQueueSize'] = 10 (needs to be int vs long)
        config['reconnectionBound'] = int(20)
        config['retain'] = True
        config['password'] = "foobar"
        config['trustStore'] = "/tmp/no-such-trustStore"
        config['trustStorePassword'] = "woohoo"
        config['keyStore'] = "/tmp/no-such-keyStore"
        config['keyStorePassword'] = "woohoo"

        # create the connector's configuration property map
        config['serverURI'] = "tcp://localhost:1883"
        config['userID'] = "user1id"
        config[' password'] = "user1passwrd"
     
        # create the connector
        mqstream = MqttStreams(topo, config)
     
        # publish a python source stream to the topic "python.topic1"
        topic = "python.topic1"
        src = topo.source(test_functions.mqtt_publish)
        mqs = mqstream.publish(src, topic)

        # subscribe to the topic "python.topic1"
        topic = ["python.topic1", ]
        mqs = mqstream.subscribe(topic)
        mqs.print()


    Configuration properties apply to publish and
    subscribe unless stated otherwise.

    serverURI
         Required String. URI to the MQTT server, either
        tcp://<hostid>[:<port>]}
        or ssl://<hostid>[:<port>]}.
        The port defaults to 1883 for "tcp:" and 8883 for "ssl:" URIs.
    
    clientID
         Optional String. A unique identifier for a connection
        to the MQTT server.
        he MQTT broker only allows a single
        onnection for a particular  clientID.
        By default a unique client ID is automatically
        generated for each use of  publish() and subscribe().
        The specified clientID is used for the first
        publish() or subscribe() use and
        suffix is added for each subsequent uses.
    
    keepAliveInterval
        Optional Integer.  Automatically generate a MQTT
        ping message to the server if a message or ping hasn't been
        sent or received in the last keelAliveInterval seconds.  
        Enables the client to detect if the server is no longer available
        without having to wait for the TCP/IP timeout.  
        A value of 0 disables keepalive processing.
        The default is 60.
    
    commandTimeout
        Optional Long. The maximum time in milliseconds
        to wait for a MQTT connect or publish action to complete.
        A value of 0 causes the client to wait indefinitely.
        The default is 0.
    
    period
        Optional Long. The time in milliseconds before
        attempting to reconnect to the server following a connection failure.
        The default is 60000.
    
    userID
        Optional String.  The identifier to use when authenticating
        with a server configured to require that form of authentication.
    
    password
        Optional String.  The identifier to use when authenticating
        with server configured to require that form of authentication.
    
    trustStore
        Optional String. The pathname to a file containing the
        public certificate of trusted MQTT servers.  If a relative path
        is specified, the path is relative to the application directory.
        Required when connecting to a MQTT server with an 
        ssl:/... serverURI.
    
    trustStorePassword
        Required String when trustStore is used.
        The password needed to access the encrypted trustStore file.
    
    keyStore
        Optional String. The pathname to a file containing the
        MQTT client's public private key certificates.
        If a relative path is specified, the path is relative to the
        application directory. 
        Required when an MQTT server is configured to use SSL client authentication.
    
    keyStorePassword
        Required String when keyStore is used.
        The password needed to access the encrypted keyStore file.

    messageQueueSize
        [subscribe] Optional Integer. The size, in number
        of messages, of the subscriber's internal receive buffer.  Received
        messages are added to the buffer prior to being converted to a
        stream tuple. The receiver blocks when the buffer is full.
        The default is 50.
    
    retain
        [publish] Optional Boolean. Indicates if messages should be
        retained on the MQTT server.  Default is false.
    
    qos
        Optional Integer. The default
         MQTT quality of service used for message handling.
        The default is 0.

    """

    def __init__(self, topology, config):
        self.topology = topology
        self.config = config.copy()
        self.opCnt = 0

    def publish(self, pub_stream, topic):
        parms = self.config.copy()
        parms['topic'] = topic
        parms['dataAttributeName'] = "string"
        if (++self.opCnt > 1):
            # each op requires its own clientID
            clientId = parms['clientID']
            if (clientId is not None and len(clientId) > 0):
                parms['clientID'] = clientId + "-" + str(id(self)) + "-" + str(self.opCnt)
        # convert pub_stream outputport schema from spl po to spl rstring type
        forOp = pub_stream._map(streamsx.topology.functions.identity, schema.CommonSchema.String)
        op = self.topology.graph.addOperator(kind="com.ibm.streamsx.messaging.mqtt::MQTTSink")
        op.addInputPort(outputPort=forOp.oport)
        op.setParameters(parms)
        return None

    def subscribe(self, topic):
        parms = self.config.copy()
        if (parms['retain'] is not None):
            del parms['retain']
        parms['topics'] = topic
        parms['topicOutAttrName'] = "topic"
        parms['dataAttributeName'] = "string"
        if (++self.opCnt > 1):
            # each op requires its own clientID
            clientId = parms['clientID']
            if (clientId is not None and len(clientId) > 0):
                parms['clientID'] = clientId + "-" + str(id(self)) + "-" + str(self.opCnt)
        op = self.topology.graph.addOperator(kind="com.ibm.streamsx.messaging.mqtt::MQTTSource")
        oport = op.addOutputPort(schema=schema.StreamSchema("tuple<rstring topic, rstring string>"))
        op.setParameters(parms)
        pop = self.topology.graph.addPassThruOperator()
        pop.addInputPort(outputPort=oport)
        pOport = pop.addOutputPort(schema=schema.CommonSchema.String)
        return Stream(self.topology, pOport)



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2017
from typing import Any, List

from streamsx.topology.topology import Topology
from streamsx.topology.schema import _AnySchema

class ExtensionOperator(object):
    def __init__(self, topology: Topology, kind: str, inputs: Any=None, schemas: List[_AnySchema]=None, params: Any=None, name: str=None) -> None: ...
    def params(self) -> Any: ...



from typing import Union
from enum import Enum
from datetime import timedelta

class ConsistentRegionConfig(object):
    class Trigger(Enum):
        OPERATOR_DRIVEN = 1
        PERIODIC = 2

    def operator_driven(drainTimeout:Union[timedelta, float]=180, resetTimeout:Union[timedelta, float]=180, maxConsecutiveAttempts:int=5) -> 'ConsistentRegionConfig': ...

    def periodic(period:Union[timedelta, float], drainTimeout:Union[timedelta, float]=180, resetTimeout:Union[timedelta, float]=180, maxConsecutiveAttempts:int=5) -> 'ConsistentRegionConfig': ...



from setuptools import setup
setup(
  name = 'streamsx',
  packages = ['streamsx', 'streamsx.spl', 'streamsx.topology', 'streamsx.scripts', 'streamsx._streams'],
  include_package_data=True,
  version = '1.11.8',
  description = 'IBM Streams Python Support',
  long_description = open('DESC.txt').read(),
  author = 'IBM Streams @ github.com',
  author_email = 'debrunne@us.ibm.com',
  license='Apache License - Version 2.0',
  url = 'https://github.com/IBMStreams/pypi.streamsx',
  download_url = 'https://github.com/IBMStreams/pypi.streamsx/tarball/1.11.8',
  keywords = ['streams', 'ibmstreams', 'streaming', 'analytics', 'streaming-analytics'],
  classifiers = [
    'Development Status :: 5 - Production/Stable',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
  install_requires=['requests', 'future', 'dill', 'enum34'],
  entry_points = {
        'console_scripts': [
            'streamsx-runner=streamsx.scripts.runner:main',
            'streams-service=streamsx.scripts.service:main',
            'streamsx-info=streamsx.scripts.info:main',
            'spl-python-extract=streamsx.scripts.extract:main'
        ],
  },
)



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2016,2017

"""
REST API bindings for IBM® Streams & Streaming Analytics service.

****************
Streams REST API
****************

The Streams REST API provides programmatic access to configuration and status information for IBM Streams objects such as domains, instances, and jobs. 

:py:class:`StreamsConnection` is the entry point to using the Streams REST API
from Python. Through its functions and the returned objects status information
can be obtained for items such as :py:class:`instances <.rest_primitives.Instance>` and :py:class:`jobs <.rest_primitives.Job>`.

****************************
Streaming Analytics REST API
****************************

You can use the Streaming Analytics REST API to manage your service instance and the IBM Streams jobs that are running on the instance. The Streaming Analytics REST API is accessible from IBM Cloud applications that are bound to your service instance or from an application outside of IBM Cloud that is configured with the service instance VCAP information.

:py:class:`StreamingAnalyticsConnection` is the entry point to using the
Streaming Analytics REST API. The function :py:func:`~StreamingAnalyticsConnection.get_streaming_analytics` returns a :py:class:`~.rest_primitives.StreamingAnalyticsService` instance which is the wrapper around the Streaming Analytics REST API. This API allows functions such as :py:meth:`start <streamsx.rest_primitives.StreamingAnalyticsService.start_instance>` and :py:meth:`stop <streamsx.rest_primitives.StreamingAnalyticsService.stop_instance>` the service instance.

In addition `StreamingAnalyticsConnection` extends from :py:class:`StreamsConnection` and thus provides access to the Streams REST API for the service instance.

.. seealso::
    `IBM Streams REST API overview <https://www.ibm.com/support/knowledgecenter/SSCRJU_4.2.0/com.ibm.streams.restapi.doc/doc/restapis.html>`_
        Reference documentation for the Streams REST API.
    `Streaming Analytics REST API <https://console.ng.bluemix.net/apidocs/220-streaming-analytics?&language=node#introduction>`_
        Reference documentation for the Streaming Analytics service REST API.

.. seealso:: :ref:`sas-main`
"""
from future.builtins import *

import os
import json
import logging
import requests
from pprint import pformat
import streamsx.topology.context

import streamsx._streams._version
__version__ = streamsx._streams._version.__version__

from streamsx import st
from .rest_primitives import (Domain, Instance, Installation, RestResource, _StreamsRestClient, StreamingAnalyticsService, _streams_delegator,
    _exact_resource, _IAMStreamsRestClient, _IAMConstants)

logger = logging.getLogger('streamsx.rest')


class StreamsConnection:
    """Creates a connection to a running distributed IBM Streams instance and exposes methods to retrieve the state of
    that instance.

    Streams maintains information regarding the state of its resources. For example, these resources could include the
    currently running Jobs, Views, PEs, Operators, and Domains. The :py:class:`StreamsConnection` provides methods to
    retrieve that information.

    Args:
        username (str): Username of an authorized Streams user. If ``None``, the username is taken from the ``STREAMS_USERNAME`` environment variable. If the ``STREAMS_USERNAME`` environment variable is not set, the default `streamsadmin` is used.

        password(str): Password for `username` If ``None``, the password is taken from the ``STREAMS_PASSWORD`` environment variable. If the ``STREAMS_PASSWORD`` environment variable is not set, the default `passw0rd` is used to match the Streams Quick Start edition setup.

        resource_url(str): Root URL for IBM Streams REST API. If ``None``, the URL is taken from the ``STREAMS_REST_URL`` environment variable. If the ``REST_URL`` environment variable is not set, then ``streamtool geturl --api`` is used to obtain the URL.

    Example:
        >>> resource_url = "https://streamsqse.localdomain:8443/streams/rest/resources"
        >>> sc = StreamsConnection("streamsadmin", "passw0rd", resource_url)
        >>> sc.session.verify=False  # manually disable SSL verification, if needed
        >>> instances = sc.get_instances()
        >>> jobs_count = 0
        >>> for instance in instances:
        >>>     jobs_count += len(instance.get_jobs())
        >>> print("There are {} jobs across all instances.".format(jobs_count))
        There are 10 jobs across all instances.

    Attributes:
        session (:py:class:`requests.Session`): Requests session object for making REST calls.
    """
    def __init__(self, username=None, password=None, resource_url=None):
        """specify username, password, and resource_url"""
        streamsx._streams._version._mismatch_check(__name__)
        if username and password:
            # resource URL can be obtained via streamtool geturl or REST call
            pass
        elif st._has_local_install:
            # Assume quickstart
            username = os.getenv("STREAMS_USERNAME", "streamsadmin")
            password = os.getenv("STREAMS_PASSWORD", "passw0rd")
        else:
            raise ValueError("Must supply either a IBM Cloud VCAP Services or a username, password"
                             " to the StreamsConnection constructor.")

        if not resource_url and 'STREAMS_REST_URL' in os.environ:
            resource_url = os.environ['STREAMS_REST_URL']
        
        self._resource_url = resource_url
        self.rest_client = _StreamsRestClient(username, password)
        self.rest_client._sc = self
        self.session = self.rest_client.session
        self._analytics_service = False
        self._delegator_impl = None
        self._domains = None

    @property
    def _delegator(self):
        if self._delegator_impl is None:
            self._delegator_impl = _streams_delegator(self)
        return self._delegator_impl

    @property
    def resource_url(self):
        """str: Root URL for IBM Streams REST API"""
        self._resource_url = self._resource_url or st.get_rest_api()
        return self._resource_url

    def _get_elements(self, resource_name, eclass, id=None):
        for resource in self.get_resources():
            if resource.name == resource_name:
                elements = []
                for json_element in resource.get_resource()[resource_name]:
                    if not _exact_resource(json_element, id):
                        continue
                    elements.append(eclass(json_element, self.rest_client))
                return elements

    def _get_element_by_id(self, resource_name, eclass, id):
        """Get a single element matching an id"""
        elements = self._get_elements(resource_name, eclass, id=id)
        if not elements:
            raise ValueError("No resource matching: {0}".format(id))
        if len(elements) == 1:
            return elements[0]
        raise ValueError("Multiple resources matching: {0}".format(id))

    def get_domains(self):
        """Retrieves available domains.

        Returns:
            :py:obj:`list` of :py:class:`~.rest_primitives.Domain`: List of available domains
        """
        # Domains are fixed and actually only one per REST api.
        if self._domains is None:
            self._domains = self._get_elements('domains', Domain)
        return self._domains

    def get_domain(self, id):
        """Retrieves available domain matching a specific domain ID

        Args:
            id (str): domain ID

        Returns:
            :py:class:`~.rest_primitives.Domain`: Domain matching `id`

        Raises:
            ValueError: No matching domain exists.
        """
        return self._get_element_by_id('domains', Domain, id)
  
    def get_instances(self):
        """Retrieves available instances.

        Returns:
            :py:obj:`list` of :py:class:`~.rest_primitives.Instance`: List of available instances
        """
        return self._get_elements('instances', Instance)

    def get_instance(self, id):
        """Retrieves available instance matching a specific instance ID.

        Args:
            id (str): Instance identifier to retrieve.

        Returns:
            :py:class:`~.rest_primitives.Instance`: Instance matching `id`.

        Raises:
            ValueError: No matching instance exists or multiple matching instances exist.
        """
        return self._get_element_by_id('instances', Instance, id)

    def get_installations(self):
        """Retrieves a list of all known Streams installations.

        Returns:
            :py:obj:`list` of :py:class:`~.rest_primitives.Installation`: List of all Installation resources.
        """
        return self._get_elements('installations', Installation)

    def get_resources(self):
        """Retrieves a list of all known Streams high-level REST resources.

        Returns:
            :py:obj:`list` of :py:class:`~.rest_primitives.RestResource`: List of all Streams high-level REST resources.
        """
        json_resources = self.rest_client.make_request(self.resource_url)['resources']
        return [RestResource(resource, self.rest_client) for resource in json_resources]

    def __str__(self):
        return pformat(self.__dict__)


class StreamingAnalyticsConnection(StreamsConnection):
    """Creates a connection to a running Streaming Analytics service and exposes methods
    to retrieve the state of the service and its instance.

    Args:
        vcap_services (str, optional): VCAP services (JSON string or a filename whose content contains a JSON string).
            If not specified, it uses the value of **VCAP_SERVICES** environment variable.
        service_name (str, optional): Name of the Streaming Analytics service.
            If not specified, it uses the value of **STREAMING_ANALYTICS_SERVICE_NAME** environment variable.

    Example:
        >>> # Assume environment variable VCAP_SERVICES has correct information
        >>> sc = StreamingAnalyticsConnection(service_name='Streaming-Analytics')
        >>> print(sc.get_streaming_analytics().get_instance_status())
        {'plan': 'Standard', 'state': 'STARTED', 'enabled': True, 'status': 'running'}

    .. seealso: :ref:`sas-access`
    """
    def __init__(self, vcap_services=None, service_name=None):
        streamsx._streams._version._mismatch_check(__name__)
        self.service_name = service_name or os.environ.get('STREAMING_ANALYTICS_SERVICE_NAME')
        self.credentials = _get_credentials(_get_vcap_services(vcap_services), self.service_name)
        self._resource_url = None

        self._iam = False
        if _IAMConstants.V2_REST_URL in self.credentials and not ('userid' in self.credentials and 'password' in self.credentials):
            self._iam = True

        if self._iam:
            self.rest_client = _IAMStreamsRestClient._create(self.credentials)
        else:
            self.rest_client = _StreamsRestClient(self.credentials['userid'], self.credentials['password'])
        self.rest_client._sc = self
        self.session = self.rest_client.session
        self._analytics_service = True
        self._sas = StreamingAnalyticsService(self.rest_client, self.credentials)
        self._delegator_impl = self._sas._delegator

    @staticmethod
    def of_definition(service_def):
       """Create a connection to a Streaming Analytics service.

       The single service is defined by `service_def` which can be one of

               * The `service credentials` copied from the `Service credentials` page of the service console (not the Streams console). Credentials are provided in JSON format. They contain such as the API key and secret, as well as connection information for the service. 
               * A JSON object (`dict`) of the form: ``{ "type": "streaming-analytics", "name": "service name", "credentials": {...} }`` with the `service credentials` as the value of the ``credentials`` key.

       Args:
           service_def(dict): Definition of the service to connect to.

       Returns:
           StreamingAnalyticsConnection: Connection to defined service.
       """
       vcap_services = streamsx.topology.context._vcap_from_service_definition(service_def)
       service_name = streamsx.topology.context._name_from_service_definition(service_def)
       return StreamingAnalyticsConnection(vcap_services, service_name)
        
    @property
    def resource_url(self):
        """str: Root URL for IBM Streams REST API"""
        if self._iam:
            self._resource_url = self._resource_url or _get_iam_rest_api_url_from_creds(self.rest_client, self.credentials)
        else:
            self._resource_url = self._resource_url or _get_rest_api_url_from_creds(self.session, self.credentials)
        return self._resource_url

    def get_streaming_analytics(self):
        """Returns a :py:class:`~.rest_primitives.StreamingAnalyticsService` to allow further interaction with
        the Streaming Analytics service.

        Returns:
            :py:class:`~.rest_primitives.StreamingAnalyticsService`:
                Object for interacting with the Streaming Analytics service.
        """
        return self._sas


def _get_vcap_services(vcap_services=None):
    """Retrieves the VCAP Services information from the `ConfigParams.VCAP_SERVICES` field in the config object. If
    `vcap_services` is not specified, it takes the information from VCAP_SERVICES environment variable.

    Args:
        vcap_services (str): Try to parse as a JSON string, otherwise, try open it as a file.
        vcap_services (dict): Return the dict as is.

    Returns:
        dict: A dict representation of the VCAP Services information.

    Raises:
        ValueError:
            * if `vcap_services` nor VCAP_SERVICES environment variable are specified.
            * cannot parse `vcap_services` as a JSON string nor as a filename.
    """
    vcap_services = vcap_services or os.environ.get('VCAP_SERVICES')
    if not vcap_services:
        raise ValueError(
            "VCAP_SERVICES information must be supplied as a parameter or as environment variable 'VCAP_SERVICES'")
    # If it was passed to config as a dict, simply return it
    if isinstance(vcap_services, dict):
        return vcap_services
    try:
        # Otherwise, if it's a string, try to load it as json
        vcap_services = json.loads(vcap_services)
    except json.JSONDecodeError:
        # If that doesn't work, attempt to open it as a file path to the json config.
        try:
            with open(vcap_services) as vcap_json_data:
                vcap_services = json.load(vcap_json_data)
        except:
            raise ValueError("VCAP_SERVICES information is not JSON or a file containing JSON:", vcap_services)
    return vcap_services


def _get_credentials(vcap_services, service_name=None):
    """Retrieves the credentials of the VCAP Service of the specified `service_name`.  If
    `service_name` is not specified, it takes the information from STREAMING_ANALYTICS_SERVICE_NAME environment
    variable.

    Args:
        vcap_services (dict): A dict representation of the VCAP Services information.
        service_name (str): One of the service name stored in `vcap_services`

    Returns:
        dict: A dict representation of the credentials.

    Raises:
        ValueError:  Cannot find `service_name` in `vcap_services`
    """
    service_name = service_name or os.environ.get('STREAMING_ANALYTICS_SERVICE_NAME', None)
    # Get the service corresponding to the SERVICE_NAME
    services = vcap_services['streaming-analytics']
    creds = None
    for service in services:
        if service['name'] == service_name:
            creds = service['credentials']
            break

    # If no corresponding service is found, error
    if creds is None:
        raise ValueError("Streaming Analytics service " + str(service_name) + " was not found in VCAP_SERVICES")
    return creds


def _get_rest_api_url_from_creds(session, credentials):
    """Retrieves the Streams REST API URL from the provided credentials.
    Args:
        session (:py:class:`requests.Session`): A Requests session object for making REST calls
        credentials (dict): A dict representation of the credentials.
    Returns:
        str: The remote Streams REST API URL.
    """
    resources_url = credentials['rest_url'] + credentials['resources_path']
    try:
        response_raw = session.get(resources_url, auth=(credentials['userid'], credentials['password']))
        response = response_raw.json()
    except:
        logger.error("Error while retrieving rest REST url from: " + resources_url)
        raise

    response_raw.raise_for_status()

    rest_api_url = response['streams_rest_url'] + '/resources'
    return rest_api_url

def _get_iam_rest_api_url_from_creds(rest_client, credentials):
    """Retrieves the Streams REST API URL from the provided credentials using iam authentication.
    Args:
        rest_client (:py:class:`rest_primitives._IAMStreamsRestClient`): A client  for making REST calls using IAM authentication
        credentials (dict): A dict representation of the credentials.
    Returns:
        str: The remote Streams REST API URL.
    """
    res = rest_client.make_request(credentials[_IAMConstants.V2_REST_URL])
    base = res['streams_self']
    end = base.find('/instances')
    return base[:end] + '/resources'
    



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2017,2018
from future.builtins import *

import inspect
import logging
import os
import pkg_resources
import sys
import streamsx
from pkgutil import extend_path

_TRACE = logging.getLogger('streamsx.runtime')

def _add_to_sys_path(dir_):
    if _TRACE.isEnabledFor(logging.DEBUG):
        _TRACE.debug('Potential addition to sys.path: %s EXISTS %s', dir_, str(os.path.isdir(dir_)))
    if os.path.isdir(dir_) and dir_ not in sys.path and os.listdir(dir_):
        _TRACE.debug('Inserting as first entry to sys.path: %s', dir_)
        sys.path.insert(0, dir_)
        pkg_resources.working_set.add_entry(dir_)
        # In case a streamsx module (e.g. streamsx.bm) 
        # is included in the additional code
        if os.path.isdir(os.path.join(dir_, 'streamsx')):
            streamsx.__path__ = extend_path(streamsx.__path__, streamsx.__name__)
        return True
    return False

def _setup_operator(tk_dir):
    pydir = os.path.join(tk_dir, 'opt', 'python')
    changed = _add_to_sys_path(os.path.join(pydir, 'modules'))
    changed = _add_to_sys_path(os.path.join(pydir, 'packages')) or changed

    if changed and _TRACE.isEnabledFor(logging.INFO):
        _TRACE.info('Updated sys.path: %s', str(sys.path))

                
class _Setup(object):
    _DONE = False

    @staticmethod
    def _setup(out_dir):
        if _Setup._DONE:
            return
        _Setup._DONE = True
        bundle_site_dir = _Setup._add_output_packages(out_dir)
        _Setup._trace_packages(bundle_site_dir)

    @staticmethod
    def _add_output_packages(out_dir):
        py_dir = _Setup._pip_base_dir(out_dir)
        vdir = 'python' + str(sys.version_info.major) + '.' + str(sys.version_info.minor)
        site_pkg = os.path.join(py_dir, 'lib', vdir, 'site-packages')
        _add_to_sys_path(site_pkg)
        return site_pkg

    @staticmethod
    def _pip_base_dir(out_dir):
        """Base Python directory for pip within the output directory"""
        return os.path.join(out_dir, 'etc', 'streamsx.topology', 'python')

    @staticmethod
    def _trace_packages(bundle_site_dir):
        if not _TRACE.isEnabledFor(logging.INFO):
            return

        _TRACE.info('sys.path: %s', str(sys.path))
        dists = list(pkg_resources.working_set)
        dists.sort(key=lambda d : d.key)
        _TRACE.info('*** Python packages ***')
        for pkg_ in dists:
            _TRACE.info(repr(pkg_))
        _TRACE.info('*** End Python packages ***')


def _setup(out_dir):
    _Setup._setup(out_dir)

# Application logic runs within an operator within a
# statement context, effectively a with statement.
# This means that for logic that has __enter__ and __exit__  methods
# (technically its type has the methods)
#
# This allows the logic to create and dispose of objects that
# cannot be pickled such as open files, custom metrics etc.
# 
# __enter__ is called:
#   a) when the operator starts up before tuple processing
#   b) when the operator resets, to inital state or from a checkpoint
#      enter is called on the new instance.
#
# __exit__ is called:
#   a) when the operator shuts down
#   b) when an exception occurs in tuple processing
#   c) when the operator resets (on the current instance)
#
#  Note: __exit__ is only called if __enter__ was called previously 
#  Note: in b) if __exit__ returns a true value then the exception is suppressed
#
# Two attributes are set in the object being wrapped to manage context:
#
# _streamsx_ec_context : Boolean indicating if the object has context methods
# _streamsx_ec_entered  : Has __enter__ been called on this instance.
#
# Note that the top-level Python object seen by the C++ primitive operator
# maintains these attributes and is responsible for calling __enter__
# on any wrapped logic.
#
# For topology:
#     The top-level object is an instance of
#     streamsx.topology.runtime._FunctionCallable wrapping the
#     application logic.
#
# For SPL primitives:
#     Source operator uses streamsx.spl.runtime._SourceIterator
#     that wraps the iterable maintains a reference to the iterator.
#
#     Otherwise the top-level object is an instance of sub-class
#     of the application logic.
#
# The C++ operator calls
#
#   _call_enter() - to enter the object into the context
#   _call_exit() - to exit the object from the context.
#
#   These methods are responsible for seeing if the underlying
#   application logic's methods are called.

def _has_context_methods(cls):
    return hasattr(cls, '__enter__') and hasattr(cls, '__exit__')

def _call_enter(obj, opc):
    if obj._streamsx_ec_context or obj._streamsx_ec_cls:
        obj._streamsx_ec_opc = opc
        if obj._streamsx_ec_context:
            obj.__enter__()
            obj._streamsx_ec_entered = True

def _call_exit(obj, exc_info=None):
    if obj._streamsx_ec_context and obj._streamsx_ec_entered:
        try:
            if exc_info is None:
                ev = obj.__exit__(None,None,None)
            else:
                exc_type = exc_info[0]
                exc_value = exc_info[1] if len(exc_info) >=2 else None
                traceback = exc_info[2] if len(exc_info) >=3 else None

                ev = obj.__exit__(exc_type, exc_value, traceback)
                if ev and exc_type is not None:
                    # Remain in the context
                    return ev
            obj._streamsx_ec_entered = False
            obj._streamsx_ec_opc = None
            return ev
        except:
            obj._streamsx_ec_entered = False
            obj._streamsx_ec_opc = None
            raise
    obj._streamsx_ec_opc = None

# A _WrappedInstance is used to wrap the functional logic
# passed into a function like map when declaring the graph.
# The wrapping occurs at topology declaration time and the
# instance of _WrappedInstance becomes the "users" logic
# that is passed in as the functional operator's parameter.
# 
# If no_context is true then it's guaranteed that
# callable_ does not have __enter__, __exit__ methods
class _WrapOpLogic(object):
    def __init__(self, callable_, no_context=None):
        self._callable = callable_

        is_cls = not inspect.isfunction(callable_)
        is_cls = is_cls and not inspect.isbuiltin(callable_)
        is_cls = is_cls and not inspect.isclass(callable_)
        self._streamsx_ec_cls = is_cls

        if is_cls and not no_context:
            if hasattr(callable_, '_streamsx_ec_context'):
                self._streamsx_ec_context = callable_._streamsx_ec_context
            else:
                self._streamsx_ec_context = streamsx._streams._runtime._has_context_methods(type(callable_))
        else:
            self._streamsx_ec_context = False

        self._streamsx_ec_entered = False

    def __enter__(self):
        if self._streamsx_ec_context or self._streamsx_ec_cls:
            self._callable._streamsx_ec_opc = self._streamsx_ec_opc
            if self._streamsx_ec_context:
                self._callable.__enter__()
    
    def __exit__(self, exc_type, exc_value, traceback):
        if self._streamsx_ec_context:
            ev = self._callable.__exit__(exc_type, exc_value, traceback)
            if not ev:
                self._streamsx_ec_opc = None
            return ev



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2017
from typing import Any, Type, Union

import enum

def is_common(schema: Any) -> bool: ...

_AnySchema = Union[StreamSchema, CommonSchema, str]

class StreamSchema(object):
    def __init__(self, schema: str) -> None: ...
    @property
    def style(self) -> Type: ...
    def as_tuple(self, named : Union[bool,str]=None) -> 'StreamSchema': ...
    def as_dict(self) -> 'StreamSchema': ...
    def extend(self, schema: Any) -> 'StreamSchema': ...


class CommonSchema(enum.Enum):
    Python: StreamSchema
    Json: StreamSchema
    String: StreamSchema
    Binary: StreamSchema
    XML: StreamSchema
    def extend(self, schema: Any) -> StreamSchema: ...

class _SchemaParser: ...



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2017,2018
"""

Testing support for streaming applications.

********
Overview
********

Allows testing of a streaming application by creation conditions
on streams that are expected to become valid during the processing.
`Tester` is designed to be used with Python's `unittest` module.

A complete application may be tested or fragments of it, for example a sub-graph can be tested
in isolation that takes input data and scores it using a model.

Supports execution of the application on
:py:const:`~streamsx.topology.context.ContextTypes.STREAMING_ANALYTICS_SERVICE`,
:py:const:`~streamsx.topology.context.ContextTypes.DISTRIBUTED`
or :py:const:`~streamsx.topology.context.ContextTypes.STANDALONE`.

A :py:class:`Tester` instance is created and associated with the :py:class:`Topology` to be tested.
Conditions are then created against streams, such as a stream must receive 10 tuples using
:py:meth:`~Tester.tuple_count`.

Here is a simple example that tests a filter correctly only passes tuples with values greater than 5::

    import unittest
    from streamsx.topology.topology import Topology
    from streamsx.topology.tester import Tester

    class TestSimpleFilter(unittest.TestCase):

        def setUp(self):
            # Sets self.test_ctxtype and self.test_config
            Tester.setup_streaming_analytics(self)

        def test_filter(self):
            # Declare the application to be tested
            topology = Topology()
            s = topology.source([5, 7, 2, 4, 9, 3, 8])
            s = s.filter(lambda x : x > 5)

            # Create tester and assign conditions
            tester = Tester(topology)
            tester.contents(s, [7, 9, 8])

            # Submit the application for test
            # If it fails an AssertionError will be raised.
           tester.test(self.test_ctxtype, self.test_config)


A stream may have any number of conditions and any number of streams may be tested.

A :py:meth:`~Tester.local_check` is supported where a method of the
unittest class is executed once the job becomes healthy. This performs
checks from the context of the Python unittest class, such as
checking external effects of the application or using the REST api to
monitor the application.

A test fails-fast if any of the following occur:
    * Any condition fails. E.g. a tuple failing a :py:meth:`~Tester.tuple_check`.
    * The :py:meth:`~Tester.local_check` (if set) raises an error.
    * The job for the test:
        * Fails to become healthy.
        * Becomes unhealthy during the test run.
        * Any processing element (PE) within the job restarts.

A test timeouts if it does not fail but its conditions do not become valid.
The timeout is not fixed as an absolute test run time, but as a time since "progress"
was made. This can allow tests to pass when healthy runs are run in a constrained
environment that slows execution. For example with a tuple count condition of ten,
progress is indicated by tuples arriving on a stream, so that as long as gaps
between tuples are within the timeout period the test remains running until ten tuples appear.

.. note:: The test timeout value is not configurable.

.. note:: The submitted job (application under test) has additional elements (streams & operators) inserted to implement the conditions. These are visible through various APIs including the Streams console raw graph view. Such elements are put into the `Tester` category.

.. warning::
    Streaming Analytics service or IBM Streams 4.2 or later is required when using `Tester`.


.. versionchanged:: 1.9 - Python 2.7 supported (except with Streaming Analytics service).

"""
from __future__ import unicode_literals
from future.builtins import *

import streamsx.ec as ec
import streamsx.topology.context as stc
import csv
import os
import unittest
import logging
import collections
import pkg_resources
import threading
from streamsx.rest import StreamsConnection
from streamsx.rest import StreamingAnalyticsConnection
from streamsx.topology.context import ConfigParams
import time
import json
import sys

import streamsx.topology.tester_runtime as sttrt

import streamsx._streams._version
__version__ = streamsx._streams._version.__version__

_logger = logging.getLogger('streamsx.topology.test')

class _TestConfig(dict):
    def __init__(self, test, entries=None):
        super(_TestConfig, self).__init__()
        self._test = test
        if entries:
            self.update(entries)

class Tester(object):
    """Testing support for a Topology.

    Allows testing of a Topology by creating conditions against the contents
    of its streams.

    Conditions may be added to a topology at any time before submission.

    If a topology is submitted directly to a context then the graph
    is not modified. This allows testing code to be inserted while
    the topology is being built, but not acted upon unless the topology
    is submitted in test mode.

    If a topology is submitted through the test method then the topology
    may be modified to include operations to ensure the conditions are met.

    .. warning::
        For future compatibility applications under test should not include intended failures that cause
        a processing element to stop or restart. Thus, currently testing is against expected application behavior.

    Args:
        topology: Topology to be tested.
    """
    def __init__(self, topology):
        self.topology = topology
        topology.tester = self
        self._conditions = {}
        self.local_check = None
        self._run_for = 0

    @staticmethod
    def _log_env(test, verbose):
        streamsx._streams._version._mismatch_check(__name__)
        if verbose:
            _logger.propogate = False
            _logger.setLevel(logging.DEBUG)
            _logger.addHandler(logging.StreamHandler())

        _logger.debug("Test:%s: PYTHONHOME=%s", test.id(), os.environ.get('PYTHONHOME', '<notset>'))
        _logger.debug("Test:%s: sys.path=%s", test.id(), sys.path)
        _logger.debug("Test:%s: tester.__file__=%s", test.id(), __file__)
        srp = pkg_resources.working_set.find(pkg_resources.Requirement.parse('streamsx'))
        if srp is None:
            _logger.debug("Test:%s: streamsx not installed.", test.id())
        else:
            _logger.debug("Test:%s: %s installed at %s.", test.id(), srp, srp.location)

    @staticmethod
    def setup_standalone(test, verbose=None):
        """
        Set up a unittest.TestCase to run tests using IBM Streams standalone mode.

        Requires a local IBM Streams install define by the ``STREAMS_INSTALL``
        environment variable. If ``STREAMS_INSTALL`` is not set, then the
        test is skipped.

        A standalone application under test will run until a condition
        fails or all the streams are finalized or when the
        :py:meth:`run_for` time (if set) elapses. 
        Applications that include infinite streams must include set a
        run for time using :py:meth:`run_for` to ensure the test completes

        Two attributes are set in the test case:

            * test_ctxtype - Context type the test will be run in.
            * test_config- Test configuration.

        Args:
            test(unittest.TestCase): Test case to be set up to run tests using Tester
            verbose(bool): If `true` then the ``streamsx.topology.test`` logger is configured at ``DEBUG`` level with output sent to standard error.

        Returns: None
        """
        if not 'STREAMS_INSTALL' in os.environ:
            raise unittest.SkipTest("Skipped due to no local IBM Streams install")
        Tester._log_env(test, verbose)
        test.test_ctxtype = stc.ContextTypes.STANDALONE
        test.test_config = _TestConfig(test)

    @staticmethod
    def get_streams_version(test):
        """ Returns IBM Streams product version string for a test.

        Returns the product version corresponding to the test's setup.
        For ``STANDALONE`` and ``DISTRIBUTED`` the product version
        corresponds to the version defined by the environment variable
        ``STREAMS_INSTALL``.

        Args:
            test(unittest.TestCase): Test case setup to run IBM Streams tests.
      
        .. versionadded: 1.11
        """
        if hasattr(test, 'test_ctxtype'):
            if test.test_ctxtype == stc.ContextTypes.STANDALONE or test.test_ctxtype == stc.ContextTypes.DISTRIBUTED:
                return Tester._get_streams_product_version()
            if test.test_ctxtype == stc.ContextTypes.STREAMING_ANALYTICS_SERVICE:
                return '4.2.0.0'
        raise ValueError('Tester has not been setup.')

    @staticmethod
    def _get_streams_product_version():
        pvf = os.path.join(os.environ['STREAMS_INSTALL'], '.product')
        vers={}
        with open(pvf, "r") as cf:
            eqc = b'=' if sys.version_info.major == 2 else '='
            reader = csv.reader(cf, delimiter=eqc, quoting=csv.QUOTE_NONE)
            for row in reader:
                vers[row[0]] = row[1]
        return vers['Version']

    @staticmethod
    def _minimum_streams_version(product_version, required_version):
        rvrmf = required_version.split('.')
        pvrmf = product_version.split('.')
        for i in range(len(rvrmf)):
            if i >= len(pvrmf):
                return False
            pi = int(pvrmf[i])
            ri = int(rvrmf[i])
            if pi < ri:
                return False
            if pi > ri:
                return True
        return True

    @staticmethod
    def minimum_streams_version(test, required_version):
        """ Checks test setup matches a minimum required IBM Streams version.

        Args:
            test(unittest.TestCase): Test case setup to run IBM Streams tests.
            required_version(str): VRMF of the minimum version the test requires. Examples are ``'4.3'``, ``4.2.4``.

        Returns:
            bool: True if the setup fulfills the minimum required version, false otherwise.

        .. versionadded: 1.11
        """
        return Tester._minimum_streams_version(Tester.get_streams_version(test), required_version)

    @staticmethod
    def require_streams_version(test, required_version):
        """Require a test has minimum IBM Streams version.
 
        Skips the test if the test's setup is not at the required
        minimum IBM Streams version.

        Args:
            test(unittest.TestCase): Test case setup to run IBM Streams tests.
            required_version(str): VRMF of the minimum version the test requires. Examples are ``'4.3'``, ``4.2.4``.

        .. versionadded: 1.11
        """
        if not Tester.minimum_streams_version(test, required_version):
            raise unittest.SkipTest("Skipped as test requires IBM Streams {0} but {1} is setup for {2}.".format(required_version, Tester.get_streams_version(test), test.test_ctxtype))

    @staticmethod
    def setup_distributed(test, verbose=None):
        """
        Set up a unittest.TestCase to run tests using IBM Streams distributed mode.

        Requires a local IBM Streams install define by the ``STREAMS_INSTALL``
        environment variable. If ``STREAMS_INSTALL`` is not set then the
        test is skipped.

        The Streams instance to use is defined by the environment variables:

         * ``STREAMS_ZKCONNECT`` - Zookeeper connection string (optional)
         * ``STREAMS_DOMAIN_ID`` - Domain identifier
         * ``STREAMS_INSTANCE_ID`` - Instance identifier

        The user used to submit and monitor the job is set by the
        optional environment variables:

         * ``STREAMS_USERNAME`` - User name defaulting to `streamsadmin`.
         * ``STREAMS_PASSWORD`` - User password defaulting to `passw0rd`.

        The defaults match the setup for testing on a IBM Streams Quick
        Start Edition (QSE) virtual machine.

        .. warning::
            ``streamtool`` is used to submit the job and requires that ``streamtool`` does not prompt for authentication.  This is achieved by using ``streamtool genkey``.

            .. seealso::
                `Generating authentication keys for IBM Streams <https://www.ibm.com/support/knowledgecenter/SSCRJU_4.2.1/com.ibm.streams.cfg.doc/doc/ibminfospherestreams-user-security-authentication-rsa.html>`_

        Two attributes are set in the test case:

             * test_ctxtype - Context type the test will be run in.
             * test_config - Test configuration.

        Args:
            test(unittest.TestCase): Test case to be set up to run tests using Tester
            verbose(bool): If `true` then the ``streamsx.topology.test`` logger is configured at ``DEBUG`` level with output sent to standard error.

        Returns: None

        """
        if not 'STREAMS_INSTALL' in os.environ:
            raise unittest.SkipTest("Skipped due to no local IBM Streams install")

        domain_instance_setup = 'STREAMS_INSTANCE_ID' in os.environ and 'STREAMS_DOMAIN_ID' in os.environ
        rest_setup = 'STREAMS_REST_URL' in os.environ

        if not domain_instance_setup and not rest_setup:
            raise unittest.SkipTest("Skipped due missing environment variables")

        Tester._log_env(test, verbose)
        test.test_ctxtype = stc.ContextTypes.DISTRIBUTED
        test.test_config = _TestConfig(test)

    @staticmethod
    def setup_streaming_analytics(test, service_name=None, force_remote_build=False, verbose=None):
        """
        Set up a unittest.TestCase to run tests using Streaming Analytics service on IBM Cloud.

        The service to use is defined by:

            * VCAP_SERVICES environment variable containing `streaming_analytics` entries.
            * service_name which defaults to the value of STREAMING_ANALYTICS_SERVICE_NAME environment variable.

        If VCAP_SERVICES is not set or a service name is not defined, then the test is skipped.

        Two attributes are set in the test case:

            * test_ctxtype - Context type the test will be run in.
            * test_config - Test configuration.

        Args:
            test(unittest.TestCase): Test case to be set up to run tests using Tester
            service_name(str): Name of Streaming Analytics service to use. Must exist as an
                entry in the VCAP services. Defaults to value of STREAMING_ANALYTICS_SERVICE_NAME environment variable.
            force_remote_build(bool): Force use of the Streaming Analytics build service. If `false` and ``STREAMS_INSTALL`` is set then a local build will be used if the local environment is suitable for the service, otherwise the Streams application bundle is built using the build service.
            verbose(bool): If `true` then the ``streamsx.topology.test`` logger is configured at ``DEBUG`` level with output sent to standard error.

        If run with Python 2 the test is skipped, only Python 3.5
        is supported with Streaming Analytics service.

        Returns: None
        """
        if sys.version_info.major == 2:
            raise unittest.SkipTest('Skipped due to running with Python 2')
        if not (sys.version_info.major == 3 and sys.version_info.minor == 5):
            raise unittest.SkipTest('Skipped as Streaming Analytics service requires Python 3.5')
        if not 'VCAP_SERVICES' in os.environ:
            raise unittest.SkipTest("Skipped due to VCAP_SERVICES environment variable not set")

        test.test_ctxtype = stc.ContextTypes.STREAMING_ANALYTICS_SERVICE
        if service_name is None:
            service_name = os.environ.get('STREAMING_ANALYTICS_SERVICE_NAME', None)
        if service_name is None:
            raise unittest.SkipTest("Skipped due to no service name supplied")

        Tester._log_env(test, verbose)
        test.test_config = _TestConfig(test, {'topology.service.name': service_name})
        if force_remote_build:
            test.test_config['topology.forceRemoteBuild'] = True

    def add_condition(self, stream, condition):
        """Add a condition to a stream.

        Conditions are normally added through :py:meth:`tuple_count`, :py:meth:`contents` or :py:meth:`tuple_check`.

        This allows an additional conditions that are implementations of :py:class:`Condition`.

        Args:
            stream(Stream): Stream to be tested.
            condition(Condition): Arbitrary condition.

        Returns:
            Stream: stream
        """
        self._conditions[condition.name] = (stream, condition)
        return stream

    def tuple_count(self, stream, count, exact=True):
        """Test that a stream contains a number of tuples.

        If `exact` is `True`, then condition becomes valid when `count`
        tuples are seen on `stream` during the test. Subsequently if additional
        tuples are seen on `stream` then the condition fails and can never
        become valid.

        If `exact` is `False`, then the condition becomes valid once `count`
        tuples are seen on `stream` and remains valid regardless of
        any additional tuples.

        Args:
            stream(Stream): Stream to be tested.
            count(int): Number of tuples expected.
            exact(bool): `True` if the stream must contain exactly `count`
                tuples, `False` if the stream must contain at least `count` tuples.

        Returns:
            Stream: stream
        """
        _logger.debug("Adding tuple count (%d) condition to stream %s.", count, stream)
        name = stream.name + '_count'
        if exact:
            cond = sttrt._TupleExactCount(count, name)
            cond._desc = "{0} stream expects tuple count equal to {1}.".format(stream.name, count)
        else:
            cond = sttrt._TupleAtLeastCount(count, name)
            cond._desc = "'{0}' stream expects tuple count of at least {1}.".format(stream.name, count)
        return self.add_condition(stream, cond)

    def contents(self, stream, expected, ordered=True):
        """Test that a stream contains the expected tuples.

        Args:
            stream(Stream): Stream to be tested.
            expected(list): Sequence of expected tuples.
            ordered(bool): True if the ordering of received tuples must match expected.

        Returns:
            Stream: stream
        """
        name = stream.name + '_contents'
        if ordered:
            cond = sttrt._StreamContents(expected, name)
            cond._desc = "'{0}' stream expects tuple ordered contents: {1}.".format(stream.name, expected)
        else:
            cond = sttrt._UnorderedStreamContents(expected, name)
            cond._desc = "'{0}' stream expects tuple unordered contents: {1}.".format(stream.name, expected)
        return self.add_condition(stream, cond)

    def resets(self, minimum_resets=10):
        """Create a condition that randomly resets consistent regions.
        The condition becomes valid when each consistent region in the
        application under test has been reset `minimum_resets` times
        by the tester.


        The resets are performed at arbitrary intervals scaled to the 
        period of the region (if it is periodically triggered).

        .. note::
             A region is reset by initiating a request though the Job Control Plane. The reset is not driven by any injected failure, such as a PE restart.

        Args:
            minimum_resets(int): Minimum number of resets for each region.

        .. versionadded:: 1.11
        """
        resetter = sttrt._Resetter(self.topology, minimum_resets=minimum_resets)
        self.add_condition(None, resetter)

    def tuple_check(self, stream, checker):
        """Check each tuple on a stream.

        For each tuple ``t`` on `stream` ``checker(t)`` is called.

        If the return evaluates to `False` then the condition fails.
        Once the condition fails it can never become valid.
        Otherwise the condition becomes or remains valid. The first
        tuple on the stream makes the condition valid if the checker
        callable evaluates to `True`.

        The condition can be combined with :py:meth:`tuple_count` with
        ``exact=False`` to test a stream map or filter with random input data.

        An example of combining `tuple_count` and `tuple_check` to test a filter followed
        by a map is working correctly across a random set of values::

            def rands():
                r = random.Random()
                while True:
                    yield r.random()

            class TestFilterMap(unittest.testCase):
            # Set up omitted

                def test_filter(self):
                    # Declare the application to be tested
                    topology = Topology()
                    r = topology.source(rands())
                    r = r.filter(lambda x : x > 0.7)
                    r = r.map(lambda x : x + 0.2)

                    # Create tester and assign conditions
                    tester = Tester(topology)
                    # Ensure at least 1000 tuples pass through the filter.
                    tester.tuple_count(r, 1000, exact=False)
                    tester.tuple_check(r, lambda x : x > 0.9)


                    # Submit the application for test
                    # If it fails an AssertionError will be raised.
                    tester.test(self.test_ctxtype, self.test_config)

        Args:
            stream(Stream): Stream to be tested.
            checker(callable): Callable that must evaluate to True for each tuple.

        """
        name = stream.name + '_check'
        cond = sttrt._TupleCheck(checker, name)
        self.topology.graph.add_dependency(checker)
        return self.add_condition(stream, cond)

    def eventual_result(self, stream, checker):
        """Test a stream reaches a known result or state.

        Creates a test condition that the tuples on a stream
        eventually reach a known result or state. Each tuple
        on `stream` results in a call to ``checker(tuple_)``.

        The return from `checker` is handled as:
            * ``None`` - The condition requires more tuples to become valid.
            * `true value` - The condition has become valid.
            * `false value` - The condition has failed. Once a condition has failed it can never become valid.

        Thus `checker` is typically stateful and allows ensuring that
        condition becomes valid from a set of input tuples. For example
        in a financial application the application under test may need
        to achieve a final known balance, but due to timings of windows the
        number of tuples required to set the final balance may be variable.

        Once the condition becomes valid any false value,
        except ``None``, returned by processing of subsequent
        tuples will cause the condition to fail.

        Returning ``None`` effectively never changes the state of the condition.

        Args:
            stream(Stream): Stream to be tested.
            checker(callable): Callable that returns evaluates the state of the stream with result to the result.
       
        .. versionadded:: 1.11
        """
        name = stream.name + '_eventual'
        cond = sttrt._EventualResult(checker, name)
        self.topology.graph.add_dependency(checker)
        return self.add_condition(stream, cond)

    def local_check(self, callable):
        """Perform local check while the application is being tested.

        A call to `callable` is made after the application under test is submitted and becomes healthy.
        The check is in the context of the Python runtime executing the unittest case,
        typically the callable is a method of the test case.

        The application remains running until all the conditions are met
        and `callable` returns. If `callable` raises an error, typically
        through an assertion method from `unittest` then the test will fail.

        Used for testing side effects of the application, typically with `STREAMING_ANALYTICS_SERVICE`
        or `DISTRIBUTED`. The callable may also use the REST api for context types that support
        it to dynamically monitor the running application.

        The callable can use `submission_result` and `streams_connection` attributes from :py:class:`Tester` instance
        to interact with the job or the running Streams instance.

        Simple example of checking the job is healthy::

            import unittest
            from streamsx.topology.topology import Topology
            from streamsx.topology.tester import Tester

            class TestLocalCheckExample(unittest.TestCase):
                def setUp(self):
                    Tester.setup_distributed(self)

                def test_job_is_healthy(self):
                    topology = Topology()
                    s = topology.source(['Hello', 'World'])

                    self.tester = Tester(topology)
                    self.tester.tuple_count(s, 2)

                    # Add the local check
                    self.tester.local_check = self.local_checks

                    # Run the test
                    self.tester.test(self.test_ctxtype, self.test_config)


                def local_checks(self):
                    job = self.tester.submission_result.job
                    self.assertEqual('healthy', job.health)

        .. warning::
            A local check must not cancel the job (application under test).

        .. warning::
            A local check is not supported in standalone mode.

        Args:
            callable: Callable object.

        """
        self.local_check = callable

    def run_for(self, duration):
        """Run the test for a minimum number of seconds.

        Creates a test wide condition that becomes `valid` when the
        application under test has been running for `duration` seconds.
        Maybe be called multiple times, the test will run as long as the maximum value provided.

        Can be used to test applications without any externally visible
        streams, or streams that do not have testable conditions. For
        example a complete application may be tested by runnning it for
        for ten minutes and use :py:meth:`local_check` to test
        any external impacts, such as messages published to a
        message queue system.

        Args:
            duration(float): Minimum number of seconds the test will run for.

        .. versionadded: 1.9
        """
        self._run_for = max(self._run_for, float(duration))

    def test(self, ctxtype, config=None, assert_on_fail=True, username=None, password=None, always_collect_logs=False):
        """Test the topology.

        Submits the topology for testing and verifies the test conditions are met and the job remained healthy through its execution.

        The submitted application (job) is monitored for the test conditions and
        will be canceled when all the conditions are valid or at least one failed.
        In addition if a local check was specified using :py:meth:`local_check` then
        that callable must complete before the job is cancelled.

        The test passes if all conditions became valid and the local check callable (if present) completed without
        raising an error.

        The test fails if the job is unhealthy, any condition fails or the local check callable (if present) raised an exception.
        In the event that the test fails when submitting to the `STREAMING_ANALYTICS_SERVICE` context, the application logs are retrieved as
        a tar file and are saved to the current working directory. The filesystem path to the application logs is saved in the
        tester's result object under the `application_logs` key, i.e. `tester.result['application_logs']`

        Args:
            ctxtype(str): Context type for submission.
            config: Configuration for submission.
            assert_on_fail(bool): True to raise an assertion if the test fails, False to return the passed status.
            username(str): **Deprecated** 
            password(str): **Deprecated**
            always_collect_logs(bool): True to always collect the console log and PE trace files of the test.

        Attributes:
            result: The result of the test. This can contain exit codes, application log paths, or other relevant test information.
            submission_result: Result of the application submission from :py:func:`~streamsx.topology.context.submit`.
            streams_connection(StreamsConnection): Connection object that can be used to interact with the REST API of
                the Streaming Analytics service or instance.

        Returns:
            bool: `True` if test passed, `False` if test failed if `assert_on_fail` is `False`.

        .. deprecated:: 1.8.3
            ``username`` and ``password`` parameters. When required for
             a distributed test use the environment variables
             ``STREAMS_USERNAME`` and ``STREAMS_PASSWORD`` to define
             the Streams user.
        """
        if config is None:
            config = {}
        config['topology.alwaysCollectLogs'] = always_collect_logs

        # Look for streamsx.testing plugins
        # Each action that plugin attached to the test is
        # called passing Tester, TestCase, context type and config
        if isinstance(config, _TestConfig):
            test_ = config._test
            actions = test_._streamsx_testing_actions if hasattr(test_, '_streamsx_testing_actions') else None
            if actions:
                for action in actions:
                    _logger.debug("Adding nose plugin action %s to topology %s.", str(action), self.topology.name)
                    action(self, test_, ctxtype, config)

        # Add the conditions into the graph as sink operators
        _logger.debug("Adding conditions to topology %s.", self.topology.name)
        for ct in self._conditions.values():
            condition = ct[1]
            stream = ct[0]
            condition._attach(stream)

        # Standalone uses --kill-after parameter.
        if self._run_for and stc.ContextTypes.STANDALONE != ctxtype:
            rfn = 'run_for_' + str(int(self._run_for)) + 's'
            run_cond = sttrt._RunFor(self._run_for, rfn)
            self.add_condition(None, run_cond)
            cond_run_time = self.topology.source(run_cond, name=rfn)
            cond_run_time.category = 'Tester'
            cond_run_time._op()._layout(hidden=True)


        _logger.debug("Starting test topology %s context %s.", self.topology.name, ctxtype)

        if stc.ContextTypes.STANDALONE == ctxtype:
            passed = self._standalone_test(config)
        elif stc.ContextTypes.DISTRIBUTED == ctxtype:
            passed = self._distributed_test(config, username, password)
        elif stc.ContextTypes.STREAMING_ANALYTICS_SERVICE == ctxtype or stc.ContextTypes.ANALYTICS_SERVICE == ctxtype:
            passed = self._streaming_analytics_test(ctxtype, config)
        else:
            raise NotImplementedError("Tester context type not implemented:", ctxtype)

        if hasattr(self, 'result') and self.result.get('conditions'):
            for cn,cnr in self.result['conditions'].items():
                c = self._conditions[cn][1]
                cdesc = cn
                if hasattr(c, '_desc'):
                    cdesc = c._desc

                if 'Fail' == cnr:
                    _logger.error("Condition: %s : %s", cnr, cdesc)
                elif 'NotValid' == cnr:
                    _logger.warning("Condition: %s : %s", cnr, cdesc)
                elif 'Valid' == cnr:
                    _logger.info("Condition: %s : %s", cnr, cdesc)
        
        if assert_on_fail:
            assert passed, "Test failed for topology: " + self.topology.name
        if passed:
            _logger.info("Test topology %s passed for context:%s", self.topology.name, ctxtype)
        else:
            _logger.error("Test topology %s failed for context:%s", self.topology.name, ctxtype)
            
        return passed

    def _standalone_test(self, config):
        """ Test using STANDALONE.
        Success is solely indicated by the process completing and returning zero.
        """
        if self._run_for:
            config = config.copy()
            config['topology.standaloneRunTime'] = self._run_for + 5.0
        sr = stc.submit(stc.ContextTypes.STANDALONE, self.topology, config)
        self.submission_result = sr
        self.result = {'passed': sr['return_code'], 'submission_result': sr}
        return sr['return_code'] == 0

    def _distributed_test(self, config, username, password):
        self.streams_connection = config.get(ConfigParams.STREAMS_CONNECTION)
        if self.streams_connection is None:
            # Supply a default StreamsConnection object with SSL verification disabled, because the default
            # streams server is not shipped with a valid SSL certificate
            self.streams_connection = StreamsConnection(username, password)
            if ConfigParams.SSL_VERIFY in config:
                self.streams_connection.session.verify = config[ConfigParams.SSL_VERIFY]
            config[ConfigParams.STREAMS_CONNECTION] = self.streams_connection
        sjr = stc.submit(stc.ContextTypes.DISTRIBUTED, self.topology, config)
        self.submission_result = sjr
        if sjr['return_code'] != 0:
            _logger.error("Failed to submit job to distributed instance.")
            return False
        return self._distributed_wait_for_result(stc.ContextTypes.DISTRIBUTED, config)


    def _streaming_analytics_test(self, ctxtype, config):
        sjr = stc.submit(ctxtype, self.topology, config)
        self.submission_result = sjr
        self.streams_connection = config.get(ConfigParams.STREAMS_CONNECTION)
        if self.streams_connection is None:
            vcap_services = config.get(ConfigParams.VCAP_SERVICES)
            service_name = config.get(ConfigParams.SERVICE_NAME)
            self.streams_connection = StreamingAnalyticsConnection(vcap_services, service_name)
        if sjr['return_code'] != 0:
            _logger.error("Failed to submit job to Streaming Analytics instance")
            return False
        return self._distributed_wait_for_result(ctxtype, config)

    def _distributed_wait_for_result(self, ctxtype, config):

        cc = _ConditionChecker(self, self.streams_connection, self.submission_result)
        # Wait for the job to be healthy before calling the local check.
        if cc._wait_for_healthy():
            self._start_local_check()
            self.result = cc._complete()
            if self.local_check is not None:
                self._local_thread.join()
        else:
            _logger.error ("Job %s Wait for healthy failed", cc._job_id)
            self.result = cc._end(False, _ConditionChecker._UNHEALTHY)

        self.result['submission_result'] = self.submission_result

        if not self.result['passed'] or config['topology.alwaysCollectLogs']:
            path = self._fetch_application_logs(ctxtype)
            self.result['application_logs'] = path

        cc._canceljob(self.result)
        if hasattr(self, 'local_check_exception') and self.local_check_exception is not None:
            raise self.local_check_exception
        return self.result['passed']

    def _fetch_application_logs(self, ctxtype):
        # Fetch the logs if submitting to a Streaming Analytics Service
        if stc.ContextTypes.STREAMING_ANALYTICS_SERVICE == ctxtype or stc.ContextTypes.ANALYTICS_SERVICE == ctxtype or stc.ContextTypes.DISTRIBUTED == ctxtype:
            application_logs = self.submission_result.job.retrieve_log_trace()
            if application_logs is not None:
                _logger.info("Application logs have been fetched to " + application_logs)
            else:
                _logger.warning("Fetching job application logs is not supported in this version of Streams.")
            return application_logs

    def _start_local_check(self):
        self.local_check_exception = None
        if self.local_check is None:
            return
        self._local_thread = threading.Thread(target=self._call_local_check)
        self._local_thread.start()

    def _call_local_check(self):
        try:
            self.local_check_value = self.local_check()
        except Exception as e:
            self.local_check_value = None
            self.local_check_exception = e

# Stop nose from seeing tha Tester.test is a test (#1266)
Tester.__test__ = False

#######################################
# Internal functions
#######################################

def _result_to_dict(passed, t):
    result = {}
    result['passed'] = passed
    result['valid'] = t[0]
    result['fail'] = t[1]
    result['progress'] = t[2]
    result['conditions'] = t[3]
    return result

class _ConditionChecker(object):
    # Return from _check_once
    # (valid, fail, progress, condition_states)
    _UNHEALTHY = (False, True, False, None)

    def __init__(self, tester, sc, sjr):
        self.tester = tester
        self._sc = sc
        self._sjr = sjr
        self._instance_id = sjr['instanceId']
        self._job_id = sjr['jobId']
        self._sequences = {}
        for cn in tester._conditions:
            self._sequences[cn] = -1
        self.delay = 1.0 
        self.timeout = 30.0
        self.waits = 0
        self.additional_checks = 2

        self.job = self._find_job()

    # Wait for job to be healthy. Returns True
    # if the job became healthy, False if not.
    def _wait_for_healthy(self):
        ok_pes = 0
        while (self.waits * self.delay) < self.timeout:
            ok_ = self._check_job_health(start=True)
            if ok_ is True:
                self.waits = 0
                return True
            if ok_ is False: # actually failed
                _logger.error ("Job %s wait for healthy actually failed", self._job_id)
                return False

            # ok_ is number of ok PEs
            if ok_ <= ok_pes:
                self.waits += 1
            else:
                # making progress so don't move towards
                # the timeout
                self.waits = 0
                ok_pes = ok_
            time.sleep(self.delay)
        else:
            _logger.error ("Job %s Timed out waiting for healthy", self._job_id)

        return self._check_job_health(verbose=True)

    def _complete(self):
        while (self.waits * self.delay) < self.timeout:
            check = self._check_once()
            if check[1]:
                return self._end(False, check)
            if check[0]:
                if self.additional_checks == 0:
                    return self._end(True, check)
                self.additional_checks -= 1
                continue
            if check[2]:
                self.waits = 0
            else:
                self.waits += 1
            time.sleep(self.delay)
        else:
            _logger.error("Job %s Timed out waiting for test to complete", self._job_id)

        return self._end(False, check)

    def _end(self, passed, check):
        result = _result_to_dict(passed, check)
        return result

    def _canceljob(self, result):
        if self.job is not None:
            self.job.cancel(force=not result['passed'])

    def _check_once(self):
        if not self._check_job_health(verbose=True):
            return _ConditionChecker._UNHEALTHY
        cms = self._get_job_metrics()
        valid = True
        progress = False
        fail = False
        condition_states = {}
        for cn in self._sequences:
            condition_states[cn] = 'NotValid'
            seq_mn = sttrt.Condition._mn('seq', cn)
            # If the metrics are missing then the operator
            # is probably still starting up, cannot be valid.
            if not seq_mn in cms:
                valid = False
                continue
            seq_m = cms[seq_mn]
            if seq_m.value != self._sequences[cn]:
                # At least one condition making progress
                progress = True
                self._sequences[cn] = seq_m.value

            fail_mn = sttrt.Condition._mn('fail', cn)
            if not fail_mn in cms:
                valid = False
                continue

            fail_m = cms[fail_mn]
            if fail_m.value != 0:
                fail = True
                condition_states[cn] = 'Fail'
                continue

            valid_mn =  sttrt.Condition._mn('valid', cn)

            if not valid_mn in cms:
                valid = False
                continue
            valid_m = cms[valid_mn]

            if valid_m.value == 0:
                valid = False
            else:
                condition_states[cn] = 'Valid'

        return valid, fail, progress, condition_states

    def _check_job_health(self, start=False, verbose=False):
        self.job.refresh()
        ok_ = self.job.health == 'healthy'
        if not ok_:
            if verbose:
                _logger.error("Job %s (%s) health:%s", self.job.name, self._job_id, self.job.health)
            if not start:
                return False
        ok_pes = 0
        pes = self.job.get_pes()
        if verbose:
            _logger.info("Job %s health:%s PE count:%d", self.job.name, self.job.health, len(pes))
        for pe in pes:
            if pe.launchCount == 0:
                if verbose:
                    _logger.warn("Job %s PE %s launch count == 0", self._job_id, pe.id)
                continue # not a test failure, but not an ok_pe either
            if pe.launchCount > 1:
                if verbose or start:
                    _logger.error("Job %s PE %s launch count > 1: %s", self._job_id, pe.id, pe.launchCount)
                return False
            if pe.health != 'healthy':
                if verbose:
                    _logger.error("Job %s PE %s health: %s", self._job_id, pe.id, pe.health)
                if not start:
                    return False
            else:
                if verbose:
                    _logger.info("Job %s PE %s health: %s", self._job_id, pe.id, pe.health)
                ok_pes += 1
        return True if ok_ else ok_pes

    def _find_job(self):
        instance = self._sc.get_instance(id=self._instance_id)
        return instance.get_job(id=self._job_id)

    def _get_job_metrics(self):
        """Fetch all the condition metrics for a job.
        We refetch the metrics each time to ensure that we don't miss
        any being added, e.g. if an operator is slow to start.
        """
        cms = {}
        for op in self.job.get_operators():
            metrics = op.get_metrics(name=sttrt.Condition._METRIC_PREFIX + '*')
            for m in metrics:
                cms[m.name] = m
        return cms



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2017
from typing import Any, List, Union
import streamsx.topology.context

class _ResourceElement(object): ...

class View(_ResourceElement):
    def __init__(self, json_view: Any, rest_client: Any) -> None: ...
    def get_domain(self) -> Domain: ...
    def get_instance(self) -> Instance: ...
    def get_job(self) -> Job: ...
    def stop_data_fetch(self) -> None: ...
    def start_data_fetch(self) -> Any: ...
    def get_view_items(self) -> Any: ...

class ViewItem(_ResourceElement): ...

class Host(_ResourceElement): ...

class Job(_ResourceElement):
    def get_views(self, name: str=None) -> List[View]: ...
    def get_domain(self) -> Domain: ...
    def get_instance(self) -> Instance: ...
    def get_hosts(self) -> List[Host]: ...
    def get_operator_connections(self) -> Any: ...
    def get_operators(self) -> List[Operator]: ...
    def get_pes(self) -> List[PE]: ...
    def get_pe_connections(self) -> List[PEConnection]: ...
    def get_resource_allocations(self) -> Any: ...
    def cancel(self, force: Any=bool) -> Union[Any, bool]: ...


class Operator(_ResourceElement):
    def get_metrics(self, name: str=None) -> List[Metric]: ...

class OperatorConnection(_ResourceElement): ...


class OperatorOutputPort(_ResourceElement): ...


class Metric(_ResourceElement): ...


class PE(_ResourceElement): ...


class PEConnection(_ResourceElement): ...


class ResourceAllocation(_ResourceElement):
    def get_resource(self) -> Resource: ...
    def get_pes(self) -> List[PE]: ...
    def get_jobs(self, name: str=None) -> List[Job]: ...

class Resource(_ResourceElement):
    def get_metrics(self, name: str=None) -> List[Metric]: ...

class ActiveService(_ResourceElement): ...


class Installation(_ResourceElement): ...


class ImportedStream(_ResourceElement): ...


class ExportedStream(_ResourceElement):
    def get_operator_output_port(self) -> OperatorOutputPort: ...


class Instance(_ResourceElement):
    def get_operators(self) -> List[Operator]: ...
    def get_operator_connections(self) -> Any: ...
    def get_pes(self) -> List[PE]: ...
    def get_pe_connections(self) -> List[PEConnection]: ...
    def get_views(self, name: str=None) -> List[View]: ...
    def get_hosts(self) -> List[Host]: ...
    def get_domain(self) -> Domain: ...
    def get_jobs(self, name: str=None) -> List[Job]: ...
    def get_job(self, id: str) -> Job: ...
    def get_imported_streams(self) -> Any: ...
    def get_exported_streams(self) -> Any: ...
    def get_active_services(self) -> Any: ...
    def get_resource_allocations(self) -> Any: ...
    def get_published_topics(self) -> List[PublishedTopic]: ...
    def upload_bundle(self, bundle:str): -> 'ApplicationBundle'
    def submit_job(self, bundle:str, job_config: streamsx.topology.context.JobConfig=None): -> Job


class ApplicationBundle(_ResourceElement):
    def submit_job(self, job_config: streamsx.topology.context.JobConfig=None): -> Job


class ResourceTag(object): ...


class ActiveVersion(object): ...


class PublishedTopic(object): ...


class Domain(_ResourceElement):
    def get_instances(self) -> List[Instance]: ...
    def get_hosts(self) -> List[Host]: ...
    def get_active_services(self) -> Any: ...
    def get_resource_allocations(self) -> Any: ...
    def get_resources(self) -> Any: ...

class RestResource(_ResourceElement):
    def get_resource(self) -> Any: ...
       
class StreamingAnalyticsService(object):
    def __init__(self, rest_client: Any, credentials: Any) -> None: ...
    def cancel_job(self, job_id: Any=None, job_name: Any=None) -> Any: ...
    def start_instance(self) -> Any: ...
    def stop_instance(self) -> Any: ...
    def get_instance_status(self) -> Any: ...
